"""
Tests for the Vault runner
"""

import logging
import os
import shutil
from pathlib import Path

import pytest
from saltfactories.utils import random_string

import salt.utils.files
from tests.conftest import FIPS_TESTRUN

# pylint: disable=unused-import
from tests.support.pytest.vault import (
    vault_container_version,
    vault_delete_secret,
    vault_environ,
    vault_write_secret,
)

log = logging.getLogger(__name__)


pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_binaries_missing("dockerd", "vault", "getent"),
    pytest.mark.usefixtures("vault_container_version"),
    pytest.mark.timeout_unless_on_windows(120),
]


@pytest.fixture(scope="class")
def pillar_state_tree(tmp_path_factory):
    _pillar_state_tree = tmp_path_factory.mktemp("pillar")
    try:
        yield _pillar_state_tree
    finally:
        shutil.rmtree(str(_pillar_state_tree), ignore_errors=True)


@pytest.fixture(scope="class")
def pillar_salt_master(salt_factories, pillar_state_tree, vault_port):
    config_defaults = {
        "pillar_roots": {"base": [str(pillar_state_tree)]},
        "open_mode": True,
        "ext_pillar": [{"vault": "path=secret/path/foo"}],
        "sdbvault": {
            "driver": "vault",
        },
        "vault": {
            "auth": {"token": "testsecret"},
            "issue": {
                "token": {
                    "params": {
                        # otherwise the tests might fail because of
                        # cached tokens (should not, because by default,
                        # the cache is valid for one session only)
                        "num_uses": 1,
                    },
                },
            },
            "policies": {
                "assign": [
                    "salt_minion",
                    "salt_minion_{minion}",
                    "salt_role_{pillar[roles]}",
                    "salt_unsafe_{grains[foo]}",
                    "extpillar_this_should_always_be_absent_{pillar[vault_sourced]}",
                    "sdb_this_should_always_be_absent_{pillar[vault_sourced_sdb]}",
                    "exe_this_should_always_be_absent_{pillar[vault_sourced_exe]}",
                ],
                "cache_time": 0,
            },
            "server": {
                "url": f"http://127.0.0.1:{vault_port}",
            },
        },
        "minion_data_cache": False,
    }
    config_overrides = {
        "fips_mode": FIPS_TESTRUN,
        "publish_signing_algorithm": (
            "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
        ),
    }
    factory = salt_factories.salt_master_daemon(
        "vault-policy-int-master-uncached",
        defaults=config_defaults,
        overrides=config_overrides,
    )
    with factory.started():
        yield factory


@pytest.fixture(scope="class")
def pillar_caching_salt_master(salt_factories, pillar_state_tree, vault_port):
    config_defaults = {
        "pillar_roots": {"base": [str(pillar_state_tree)]},
        "open_mode": True,
        "ext_pillar": [{"vault": "path=secret/path/foo"}],
        "vault": {
            "auth": {"token": "testsecret"},
            "issue": {
                "token": {
                    "params": {
                        # otherwise the tests might fail because of
                        # cached tokens
                        "num_uses": 1,
                    },
                },
            },
            "policies": {
                "assign": [
                    "salt_minion",
                    "salt_minion_{minion}",
                    "salt_role_{pillar[roles]}",
                    "salt_unsafe_{grains[foo]}",
                    "extpillar_this_will_not_always_be_absent_{pillar[vault_sourced]}",
                ],
                "cache_time": 0,
            },
            "server": {
                "url": f"http://127.0.0.1:{vault_port}",
            },
        },
        "minion_data_cache": True,
    }
    config_overrides = {
        "fips_mode": FIPS_TESTRUN,
        "publish_signing_algorithm": (
            "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
        ),
    }
    factory = salt_factories.salt_master_daemon(
        "vault-policy-int-master-cached",
        defaults=config_defaults,
        overrides=config_overrides,
    )
    with factory.started():
        yield factory


@pytest.fixture(scope="class")
def pillar_salt_minion(pillar_salt_master):
    assert pillar_salt_master.is_running()
    factory = pillar_salt_master.salt_minion_daemon(
        "vault-policy-int-minion-uncached-1",
        defaults={"open_mode": True, "grains": {"foo": "bar"}},
        overrides={
            "fips_mode": FIPS_TESTRUN,
            "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
            "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
        },
    )
    with factory.started():
        # Sync All
        salt_call_cli = factory.salt_call_cli()
        ret = salt_call_cli.run("saltutil.sync_all", _timeout=120)
        assert ret.returncode == 0, ret
        yield factory


@pytest.fixture(scope="class")
def pillar_caching_salt_minion(pillar_caching_salt_master):
    assert pillar_caching_salt_master.is_running()
    factory = pillar_caching_salt_master.salt_minion_daemon(
        "vault-policy-int-minion-cached-1",
        defaults={"open_mode": True, "grains": {"foo": "bar"}},
        overrides={
            "fips_mode": FIPS_TESTRUN,
            "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
            "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
        },
    )
    with factory.started():
        # Sync All
        salt_call_cli = factory.salt_call_cli()
        ret = salt_call_cli.run("saltutil.sync_all", _timeout=120)
        assert ret.returncode == 0, ret
        yield factory


@pytest.fixture(scope="class")
def pillar_salt_run_cli(pillar_salt_master):
    return pillar_salt_master.salt_run_cli()


@pytest.fixture(scope="class")
def pillar_caching_salt_run_cli(pillar_caching_salt_master):
    return pillar_caching_salt_master.salt_run_cli()


@pytest.fixture(scope="class")
def pillar_salt_call_cli(pillar_salt_minion):
    return pillar_salt_minion.salt_call_cli()


@pytest.fixture(scope="class")
def pillar_caching_salt_call_cli(pillar_caching_salt_minion):
    return pillar_caching_salt_minion.salt_call_cli()


@pytest.fixture(scope="class")
def vault_pillar_values_policy(vault_container_version):
    vault_write_secret("secret/path/foo", vault_sourced="fail")
    try:
        yield
    finally:
        vault_delete_secret("secret/path/foo")


@pytest.mark.usefixtures("vault_pillar_values_policy")
@pytest.mark.parametrize("vault_container_version", ["latest"], indirect=True)
class TestVaultPillarPolicyTemplatesWithoutCache:
    @pytest.fixture(autouse=True)
    def pillar_policy_tree(
        self,
        pillar_salt_master,
        pillar_salt_minion,
    ):
        top_pillar_contents = f"""
        base:
          '{pillar_salt_minion.id}':
            - roles
        """
        roles_pillar_contents = """
        roles:
          - minion
          - web
        """
        top_file = pillar_salt_master.pillar_tree.base.temp_file(
            "top.sls", top_pillar_contents
        )
        roles_file = pillar_salt_master.pillar_tree.base.temp_file(
            "roles.sls", roles_pillar_contents
        )

        with top_file, roles_file:
            yield

    @pytest.fixture
    def pillar_exe_loop(self, pillar_state_tree, pillar_salt_minion):
        top_file = f"""
        base:
          '{pillar_salt_minion.id}':
            - roles
            - exe_loop
        """
        exe_loop_pillar = r"""
        vault_sourced_exe: {{ salt["vault.read_secret"]("secret/path/foo", "vault_sourced") }}
        """
        top_tempfile = pytest.helpers.temp_file("top.sls", top_file, pillar_state_tree)
        exe_loop_tempfile = pytest.helpers.temp_file(
            "exe_loop.sls", exe_loop_pillar, pillar_state_tree
        )

        with top_tempfile, exe_loop_tempfile:
            yield

    @pytest.fixture
    def pillar_sdb_loop(self, pillar_state_tree, pillar_salt_minion):
        top_file = f"""
        base:
          '{pillar_salt_minion.id}':
            - roles
            - sdb_loop
        """
        sdb_loop_pillar = r"""
        vault_sourced_sdb: {{ salt["sdb.get"]("sdb://sdbvault/secret/path/foo/vault_sourced") }}
        """
        top_tempfile = pytest.helpers.temp_file("top.sls", top_file, pillar_state_tree)
        sdb_loop_tempfile = pytest.helpers.temp_file(
            "sdb_loop.sls", sdb_loop_pillar, pillar_state_tree
        )

        with top_tempfile, sdb_loop_tempfile:
            yield

    @pytest.fixture(autouse=True)
    def minion_data_cache_absent(self, pillar_salt_run_cli, pillar_salt_minion):
        ret = pillar_salt_run_cli.run(
            "cache.flush", f"minions/{pillar_salt_minion.id}", "data"
        )
        assert ret.returncode == 0
        cached = pillar_salt_run_cli.run(
            "cache.fetch", f"minions/{pillar_salt_minion.id}", "data"
        )
        assert cached.returncode == 0
        assert not cached.data
        yield

    def test_show_policies(self, pillar_salt_run_cli, pillar_salt_minion):
        """
        Test that pillar data is refreshed correctly before rendering policies when necessary.
        This test includes the prevention of loop exceptions by the ext_pillar module
        This refresh does not include grains and pillar data targeted by these grains (unsafe anyways!).
        """
        ret = pillar_salt_run_cli.run(
            "vault.show_policies", pillar_salt_minion.id, expire=0
        )
        assert ret.data == [
            "salt_minion",
            f"salt_minion_{pillar_salt_minion.id}",
            "salt_role_minion",
            "salt_role_web",
        ]
        assert "Pillar render error: Failed to load ext_pillar vault" not in ret.stderr

    def test_show_policies_uncached_data_no_pillar_refresh(
        self, pillar_salt_run_cli, pillar_salt_minion
    ):
        """
        Test that the pillar is not refreshed when explicitly disabled
        """
        ret = pillar_salt_run_cli.run(
            "vault.show_policies", pillar_salt_minion.id, refresh_pillar=False, expire=0
        )
        assert ret.data == ["salt_minion", f"salt_minion_{pillar_salt_minion.id}"]

    @pytest.mark.skipif(
        FIPS_TESTRUN, reason="Signing with SHA1 not supported in FIPS mode."
    )
    @pytest.mark.usefixtures("pillar_exe_loop")
    def test_policy_compilation_prevents_loop_for_execution_module(
        self,
        pillar_salt_run_cli,
        pillar_salt_minion,
    ):
        """
        Test that the runner prevents a recursive cycle from happening
        """
        ret = pillar_salt_run_cli.run(
            "vault.show_policies", pillar_salt_minion.id, refresh_pillar=True, expire=0
        )
        assert ret.data == [
            "salt_minion",
            f"salt_minion_{pillar_salt_minion.id}",
            "salt_role_minion",
            "salt_role_web",
        ]
        assert "Pillar render error: Rendering SLS 'exe_loop' failed" in ret.stderr
        assert "Cyclic dependency detected while refreshing pillar" in ret.stderr
        assert "RecursionError" not in ret.stderr

    @pytest.mark.skipif(
        FIPS_TESTRUN, reason="Signing with SHA1 not supported in FIPS mode."
    )
    @pytest.mark.usefixtures("pillar_sdb_loop")
    def test_policy_compilation_prevents_loop_for_sdb_module(
        self,
        pillar_salt_run_cli,
        pillar_salt_minion,
    ):
        """
        Test that the runner prevents a recursive cycle from happening
        """
        ret = pillar_salt_run_cli.run(
            "vault.show_policies", pillar_salt_minion.id, refresh_pillar=True, expire=0
        )
        assert ret.data == [
            "salt_minion",
            f"salt_minion_{pillar_salt_minion.id}",
            "salt_role_minion",
            "salt_role_web",
        ]
        assert "Pillar render error: Rendering SLS 'sdb_loop' failed" in ret.stderr
        assert "Cyclic dependency detected while refreshing pillar" in ret.stderr
        assert "RecursionError" not in ret.stderr


@pytest.mark.usefixtures("vault_pillar_values_policy")
@pytest.mark.parametrize("vault_container_version", ["latest"], indirect=True)
class TestVaultPillarPolicyTemplatesWithCache:
    @pytest.fixture(autouse=True)
    def pillar_caching_policy_tree(
        self, pillar_caching_salt_master, pillar_caching_salt_minion
    ):
        top_pillar_contents = f"""
        base:
          '{pillar_caching_salt_minion.id}':
            - roles
        """
        roles_pillar_contents = """
        roles:
          - minion
          - web
        """
        top_file = pillar_caching_salt_master.pillar_tree.base.temp_file(
            "top.sls", top_pillar_contents
        )
        roles_file = pillar_caching_salt_master.pillar_tree.base.temp_file(
            "roles.sls", roles_pillar_contents
        )

        with top_file, roles_file:
            yield

    @pytest.fixture(autouse=True)
    def minion_data_cache_present(
        self,
        pillar_caching_salt_call_cli,
        pillar_caching_policy_tree,
    ):
        ret = pillar_caching_salt_call_cli.run("saltutil.refresh_pillar", wait=True)
        assert ret.returncode == 0
        assert ret.data is True
        yield

    @pytest.fixture(autouse=True)
    def minion_data_cache_outdated(
        self,
        minion_data_cache_present,
        pillar_caching_salt_run_cli,
        pillar_caching_salt_master,
        pillar_caching_salt_minion,
    ):
        roles_pillar_new_contents = """
        roles:
          - minion
          - web
          - fresh
        """
        roles_file = pillar_caching_salt_master.pillar_tree.base.temp_file(
            "roles.sls", roles_pillar_new_contents
        )

        cached = pillar_caching_salt_run_cli.run(
            "cache.fetch", f"minions/{pillar_caching_salt_minion.id}", "data"
        )
        assert cached.returncode == 0
        assert cached.data
        assert "pillar" in cached.data
        assert "grains" in cached.data
        assert "roles" in cached.data["pillar"]
        assert cached.data["pillar"]["roles"] == ["minion", "web"]
        with roles_file:
            yield

    def test_show_policies_cached_data_no_pillar_refresh(
        self,
        pillar_caching_salt_run_cli,
        pillar_caching_salt_minion,
    ):
        """
        Test that pillar data from cache is used when it is available
        """
        ret = pillar_caching_salt_run_cli.run(
            "vault.show_policies", pillar_caching_salt_minion.id, expire=0
        )
        assert ret.data == [
            "salt_minion",
            f"salt_minion_{pillar_caching_salt_minion.id}",
            "salt_role_minion",
            "salt_role_web",
            "salt_unsafe_bar",
            "extpillar_this_will_not_always_be_absent_fail",
        ]

    def test_show_policies_refresh_pillar(
        self,
        pillar_caching_salt_run_cli,
        pillar_caching_salt_minion,
    ):
        """
        Test that pillar data is always refreshed when requested.
        """
        ret = pillar_caching_salt_run_cli.run(
            "vault.show_policies",
            pillar_caching_salt_minion.id,
            refresh_pillar=True,
            expire=0,
        )
        assert ret.data == [
            "salt_minion",
            f"salt_minion_{pillar_caching_salt_minion.id}",
            "salt_role_minion",
            "salt_role_web",
            "salt_role_fresh",
            "salt_unsafe_bar",
        ]


# The tests above use different fixtures because I could not
# make them behave as expected otherwise.


@pytest.fixture(scope="class")
def vault_salt_master(
    salt_factories, pillar_state_tree, vault_port, vault_master_config
):
    factory = salt_factories.salt_master_daemon(
        "vault-master", defaults=vault_master_config
    )
    with factory.started():
        yield factory


@pytest.fixture(scope="class")
def vault_salt_minion(vault_salt_master):
    assert vault_salt_master.is_running()
    factory = vault_salt_master.salt_minion_daemon(
        random_string("vault-minion", uppercase=False),
        defaults={"open_mode": True, "grains": {}},
    )
    with factory.started():
        # Sync All
        salt_call_cli = factory.salt_call_cli()
        ret = salt_call_cli.run("saltutil.sync_all", _timeout=120)
        assert ret.returncode == 0, ret
        yield factory


@pytest.fixture(scope="class")
def overriding_vault_salt_minion(vault_salt_master, issue_overrides):
    assert vault_salt_master.is_running()
    factory = vault_salt_master.salt_minion_daemon(
        random_string("vault-minion", uppercase=False),
        defaults={"open_mode": True, "grains": {}},
        overrides={"vault": {"issue_params": issue_overrides}},
    )
    with factory.started():
        # Sync All
        salt_call_cli = factory.salt_call_cli()
        ret = salt_call_cli.run("saltutil.sync_all", _timeout=120)
        assert ret.returncode == 0, ret
        yield factory


@pytest.fixture(scope="class")
def vault_salt_run_cli(vault_salt_master):
    return vault_salt_master.salt_run_cli()


@pytest.fixture(scope="class")
def vault_salt_call_cli(vault_salt_minion):
    return vault_salt_minion.salt_call_cli()


@pytest.fixture(scope="class")
def pillar_roles_tree(
    vault_salt_master,
    vault_salt_minion,
):
    top_pillar_contents = f"""
    base:
      '{vault_salt_minion.id}':
        - roles
    """
    roles_pillar_contents = """
    roles:
      - dev
      - web
    # this is for entity metadata since lists are cumbersome at best
    role: foo
    """
    top_file = vault_salt_master.pillar_tree.base.temp_file(
        "top.sls", top_pillar_contents
    )
    roles_file = vault_salt_master.pillar_tree.base.temp_file(
        "roles.sls", roles_pillar_contents
    )

    with top_file, roles_file:
        yield


@pytest.fixture(scope="class")
def vault_pillar_values_approle(vault_salt_minion):
    vault_write_secret(
        f"salt/minions/{vault_salt_minion.id}", minion_id_acl_template="worked"
    )
    vault_write_secret("salt/roles/foo", pillar_role_acl_template="worked")
    try:
        yield
    finally:
        vault_delete_secret(f"salt/minions/{vault_salt_minion.id}")
        vault_delete_secret("salt/roles/foo")


@pytest.fixture(scope="class")
def vault_testing_values(vault_container_version):
    vault_write_secret("secret/path/foo", success="yeehaaw")
    try:
        yield
    finally:
        vault_delete_secret("secret/path/foo")


@pytest.fixture
def minion_conn_cachedir(vault_salt_call_cli):
    ret = vault_salt_call_cli.run("config.get", "cachedir")
    assert ret.returncode == 0
    assert ret.data
    cachedir = Path(ret.data) / "vault" / "connection"
    if not cachedir.exists():
        cachedir.mkdir(parents=True)
    yield cachedir


@pytest.fixture
def missing_auth_cache(minion_conn_cachedir):
    token_cachefile = minion_conn_cachedir / "session" / "__token.p"
    secret_id_cachefile = minion_conn_cachedir / "secret_id.p"
    for file in [secret_id_cachefile, token_cachefile]:
        if file.exists():
            file.unlink()
    yield


@pytest.fixture(scope="class")
def minion_data_cache_present(
    vault_salt_call_cli,
    vault_salt_minion,
    pillar_roles_tree,
    vault_salt_run_cli,
):
    ret = vault_salt_run_cli.run("pillar.show_top", minion=vault_salt_minion.id)
    assert ret.returncode == 0
    assert ret.data
    ret = vault_salt_call_cli.run("saltutil.refresh_pillar", wait=True)
    assert ret.returncode == 0
    assert ret.data is True
    ret = vault_salt_call_cli.run("pillar.items")
    assert ret.returncode == 0
    assert ret.data
    assert "role" in ret.data
    assert "roles" in ret.data
    yield


@pytest.fixture
def conn_cache_absent(minion_conn_cachedir):
    shutil.rmtree(minion_conn_cachedir)
    assert not minion_conn_cachedir.exists()
    yield


@pytest.fixture(scope="class")
def approles_synced(
    vault_salt_run_cli,
    minion_data_cache_present,
    vault_salt_minion,
):
    ret = vault_salt_run_cli.run("vault.sync_approles", vault_salt_minion.id)
    assert ret.returncode == 0
    assert ret.data is True
    ret = vault_salt_run_cli.run("vault.list_approles")
    assert ret.returncode == 0
    assert vault_salt_minion.id in ret.data
    yield


@pytest.fixture(scope="class")
def entities_synced(
    vault_salt_run_cli,
    minion_data_cache_present,
    vault_salt_minion,
):
    ret = vault_salt_run_cli.run("vault.sync_entities", vault_salt_minion.id)
    assert ret.returncode == 0
    assert ret.data is True
    ret = vault_salt_run_cli.run("vault.list_approles")
    assert ret.returncode == 0
    assert vault_salt_minion.id in ret.data
    ret = vault_salt_run_cli.run("vault.list_entities")
    assert ret.returncode == 0
    assert f"salt_minion_{vault_salt_minion.id}" in ret.data
    ret = vault_salt_run_cli.run("vault.show_entity", vault_salt_minion.id)
    assert ret.returncode == 0
    assert ret.data == {"minion-id": vault_salt_minion.id, "role": "foo"}
    yield


@pytest.mark.usefixtures(
    "vault_pillar_values_approle",
    "vault_testing_values",
    "pillar_roles_tree",
    "minion_data_cache_present",
)
@pytest.mark.parametrize("vault_container_version", ["latest"], indirect=True)
class TestAppRoleIssuance:
    @pytest.fixture(scope="class")
    def vault_master_config(self, pillar_state_tree, vault_port):
        return {
            "pillar_roots": {"base": [str(pillar_state_tree)]},
            "open_mode": True,
            # ensure approles/entities are generated during pillar rendering
            "ext_pillar": [
                {"vault": "path=salt/minions/{minion}"},
                {"vault": "path=salt/roles/{pillar[role]}"},
            ],
            "peer_run": {
                ".*": [
                    "vault.get_config",
                    # for test_auth_method_switch_does_not_break_minion_auth
                    "vault.generate_new_token",
                    "vault.generate_secret_id",
                ],
            },
            "vault": {
                "auth": {"token": "testsecret"},
                "cache": {
                    "backend": "file",
                },
                "issue": {
                    "allow_minion_override_params": True,
                    "type": "approle",
                    "approle": {
                        "params": {
                            "secret_id_num_uses": 0,
                            "secret_id_ttl": 1800,
                            "token_explicit_max_ttl": 1800,
                            "token_num_uses": 0,
                        }
                    },
                },
                "metadata": {
                    "entity": {
                        "minion-id": "{minion}",
                        "role": "{pillar[role]}",
                    },
                },
                "policies": {
                    "assign": [
                        "salt_minion",
                        "salt_minion_{minion}",
                        "salt_role_{pillar[roles]}",
                    ],
                },
                "server": {
                    "url": f"http://127.0.0.1:{vault_port}",
                },
            },
        }

    @pytest.fixture(scope="class")
    def issue_overrides(self):
        return {
            "token_explicit_max_ttl": 1337,
            "token_num_uses": 42,
            "secret_id_num_uses": 3,
            "secret_id_ttl": 1338,
        }

    @pytest.fixture
    def cache_auth_outdated(self, missing_auth_cache, minion_conn_cachedir, vault_port):
        vault_url = f"http://127.0.0.1:{vault_port}"
        config_data = b"\xdf\x00\x00\x00\x03\xa4auth\xdf\x00\x00\x00\x04\xadapprole_mount\xa7approle\xacapprole_name\xbavault-approle-int-minion-1\xa6method\xa5token\xa9secret_id\xc0\xa5cache\xdf\x00\x00\x00\x03\xa7backend\xa4disk\xa6config\xcd\x0e\x10\xa6secret\xa3ttl\xa6server\xdf\x00\x00\x00\x03\xa9namespace\xc0\xa6verify\xc0\xa3url"
        config_data += (len(vault_url) + 160).to_bytes(1, "big") + vault_url.encode()
        config_cachefile = minion_conn_cachedir / "config.p"
        with salt.utils.files.fopen(config_cachefile, "wb") as f:
            f.write(config_data)
        try:
            yield
        finally:
            if config_cachefile.exists():
                config_cachefile.unlink()

    @pytest.fixture
    def cache_server_outdated(self, missing_auth_cache, minion_conn_cachedir):
        config_data = b"\xdf\x00\x00\x00\x03\xa4auth\xdf\x00\x00\x00\x05\xadapprole_mount\xa7approle\xacapprole_name\xbavault-approle-int-minion-1\xa6method\xa7approle\xa7role_id\xactest-role-id\xa9secret_id\xc3\xa5cache\xdf\x00\x00\x00\x03\xa7backend\xa4disk\xa6config\xcd\x0e\x10\xa6secret\xa3ttl\xa6server\xdf\x00\x00\x00\x03\xa9namespace\xc0\xa6verify\xc0\xa3url\xb2http://127.0.0.1:8"
        config_cachefile = minion_conn_cachedir / "config.p"
        with salt.utils.files.fopen(config_cachefile, "wb") as f:
            f.write(config_data)
        try:
            yield
        finally:
            if config_cachefile.exists():
                config_cachefile.unlink()

    @pytest.mark.usefixtures("conn_cache_absent")
    def test_minion_can_authenticate(self, vault_salt_call_cli):
        """
        Test that the minion can run queries against Vault.
        The master impersonating the minion is already tested in the fixture setup
        (ext_pillar).
        """
        ret = vault_salt_call_cli.run("vault.read_secret", "secret/path/foo")
        assert ret.returncode == 0
        assert ret.data
        assert ret.data.get("success") == "yeehaaw"

    @pytest.mark.usefixtures("entities_synced")
    def test_minion_pillar_is_populated_as_expected(self, vault_salt_call_cli):
        """
        Test that ext_pillar pillar-templated paths are resolved as expectd
        (and that the ACL policy templates work on the Vault side).
        """
        ret = vault_salt_call_cli.run("pillar.items")
        assert ret.returncode == 0
        assert ret.data
        assert ret.data.get("minion_id_acl_template") == "worked"
        assert ret.data.get("pillar_role_acl_template") == "worked"

    @pytest.mark.usefixtures("approles_synced")
    @pytest.mark.usefixtures("conn_cache_absent")
    def test_minion_token_policies_are_assigned_as_expected(
        self, vault_salt_call_cli, vault_salt_minion
    ):
        """
        Test that issued tokens have the expected policies.
        """
        ret = vault_salt_call_cli.run("vault.query", "GET", "auth/token/lookup-self")
        assert ret.returncode == 0
        assert ret.data
        assert set(ret.data["data"]["policies"]) == {
            "default",
            "salt_minion",
            f"salt_minion_{vault_salt_minion.id}",
            "salt_role_dev",
            "salt_role_web",
        }

    @pytest.mark.usefixtures("cache_auth_outdated")
    def test_auth_method_switch_does_not_break_minion_auth(
        self, vault_salt_call_cli, caplog
    ):
        """
        Test that after a master configuration switch from another authentication method,
        minions with cached configuration flush it and request a new one.
        """
        ret = vault_salt_call_cli.run("vault.read_secret", "secret/path/foo")
        assert ret.returncode == 0
        assert ret.data
        assert ret.data.get("success") == "yeehaaw"
        assert "Master returned error and requested cache expiration" in caplog.text

    @pytest.mark.usefixtures("cache_server_outdated")
    def test_server_switch_does_not_break_minion_auth(
        self, vault_salt_call_cli, caplog
    ):
        """
        Test that after a master configuration switch to another server URL,
        minions with cached configuration detect the mismatch and request a
        new configuration.
        """
        ret = vault_salt_call_cli.run("vault.read_secret", "secret/path/foo")
        assert ret.returncode == 0
        assert ret.data
        assert ret.data.get("success") == "yeehaaw"
        assert "Mismatch of cached and reported server data detected" in caplog.text

    @pytest.mark.parametrize("ckey", ["config", "__token", "secret_id"])
    def test_cache_is_used_on_the_minion(
        self, ckey, vault_salt_call_cli, minion_conn_cachedir
    ):
        """
        Test that remote configuration, tokens acquired by authenticating with an AppRole
        and issued secret IDs are written to cache.
        """
        cache = minion_conn_cachedir
        if ckey == "__token":
            cache = cache / "session"
            if not cache.exists():
                cache.mkdir()
        if f"{ckey}.p" not in os.listdir(cache):
            ret = vault_salt_call_cli.run("vault.read_secret", "secret/path/foo")
            assert ret.returncode == 0
        assert f"{ckey}.p" in os.listdir(cache)

    @pytest.mark.parametrize("ckey", ["config", "__token", "secret_id"])
    def test_cache_is_used_on_the_impersonating_master(
        self, ckey, vault_salt_run_cli, vault_salt_minion
    ):
        """
        Test that remote configuration, tokens acquired by authenticating with an AppRole
        and issued secret IDs are written to cache when a master is impersonating
        a minion during pillar rendering.
        """
        cbank = f"minions/{vault_salt_minion.id}/vault/connection"
        if ckey == "__token":
            cbank += "/session"
        ret = vault_salt_run_cli.run("cache.list", cbank)
        assert ret.returncode == 0
        assert ret.data
        assert ckey in ret.data

    def test_cache_is_used_for_master_token_information(self, vault_salt_run_cli):
        """
        Test that a locally configured token is cached, including meta information.
        """
        ret = vault_salt_run_cli.run("cache.list", "vault/connection/session")
        assert ret.returncode == 0
        assert ret.data
        assert "__token" in ret.data

    @pytest.mark.usefixtures("approles_synced")
    def test_issue_param_overrides_work(
        self, overriding_vault_salt_minion, issue_overrides, vault_salt_run_cli
    ):
        """
        Test that minion overrides of issue params work for AppRoles.
        """
        ret = overriding_vault_salt_minion.salt_call_cli().run(
            "vault.query", "GET", "auth/token/lookup-self"
        )
        assert ret.returncode == 0
        assert ret.data
        ret = vault_salt_run_cli.run(
            "vault.show_approle", overriding_vault_salt_minion.id
        )
        assert ret.returncode == 0
        assert ret.data
        for val in [
            "token_explicit_max_ttl",
            "token_num_uses",
            "secret_id_num_uses",
            "secret_id_ttl",
        ]:
            assert ret.data[val] == issue_overrides[val]

    def test_impersonating_master_does_not_override_issue_param_overrides(
        self, overriding_vault_salt_minion, vault_salt_run_cli, issue_overrides
    ):
        """
        Test that rendering the pillar does not remove issue param overrides
        requested by a minion
        """
        # ensure the minion requests a new configuration
        ret = overriding_vault_salt_minion.salt_call_cli().run(
            "vault.clear_token_cache"
        )
        assert ret.returncode == 0
        # check that the overrides are applied
        ret = overriding_vault_salt_minion.salt_call_cli().run(
            "vault.query", "GET", "auth/token/lookup-self"
        )
        assert ret.returncode == 0
        assert ret.data
        assert (
            ret.data["data"]["explicit_max_ttl"]
            == issue_overrides["token_explicit_max_ttl"]
        )
        # ensure the master does not have cached authentication
        ret = vault_salt_run_cli.run("vault.clear_cache")
        assert ret.returncode == 0
        ret = vault_salt_run_cli.run(
            "pillar.show_pillar", overriding_vault_salt_minion.id
        )
        assert ret.returncode == 0
        # check that issue overrides are still present
        ret = vault_salt_run_cli.run(
            "vault.show_approle", overriding_vault_salt_minion.id
        )
        assert ret.returncode == 0
        assert ret.data
        assert (
            ret.data["token_explicit_max_ttl"]
            == issue_overrides["token_explicit_max_ttl"]
        )


@pytest.mark.usefixtures(
    "vault_testing_values", "pillar_roles_tree", "minion_data_cache_present"
)
class TestTokenIssuance:
    @pytest.fixture(scope="class")
    def vault_master_config(self, pillar_state_tree, vault_port):
        return {
            "pillar_roots": {"base": [str(pillar_state_tree)]},
            "open_mode": True,
            "ext_pillar": [{"vault": "path=secret/path/foo"}],
            "peer_run": {
                ".*": [
                    "vault.get_config",
                    "vault.generate_new_token",
                    # for test_auth_method_switch_does_not_break_minion_auth
                    "vault.generate_secret_id",
                ],
            },
            "vault": {
                "auth": {"token": "testsecret"},
                "cache": {
                    "backend": "file",
                },
                "issue": {
                    "type": "token",
                    "token": {
                        "params": {
                            "num_uses": 0,
                        }
                    },
                },
                "policies": {
                    "assign": [
                        "salt_minion",
                        "salt_minion_{minion}",
                        "salt_role_{pillar[roles]}",
                    ],
                    "cache_time": 0,
                },
                "server": {
                    "url": f"http://127.0.0.1:{vault_port}",
                },
            },
            "minion_data_cache": True,
        }

    @pytest.fixture
    def cache_auth_outdated(self, missing_auth_cache, minion_conn_cachedir, vault_port):
        vault_url = f"http://127.0.0.1:{vault_port}"
        config_data = b"\xdf\x00\x00\x00\x03\xa4auth\xdf\x00\x00\x00\x05\xadapprole_mount\xa7approle\xacapprole_name\xbavault-approle-int-minion-1\xa6method\xa7approle\xa7role_id\xactest-role-id\xa9secret_id\xc3\xa5cache\xdf\x00\x00\x00\x03\xa7backend\xa4disk\xa6config\xcd\x0e\x10\xa6secret\xa3ttl\xa6server\xdf\x00\x00\x00\x03\xa9namespace\xc0\xa6verify\xc0\xa3url"
        config_data += (len(vault_url) + 160).to_bytes(1, "big") + vault_url.encode()
        config_cachefile = minion_conn_cachedir / "config.p"
        with salt.utils.files.fopen(config_cachefile, "wb") as f:
            f.write(config_data)
        try:
            yield
        finally:
            if config_cachefile.exists():
                config_cachefile.unlink()

    @pytest.fixture(scope="class")
    def issue_overrides(self):
        # only explicit_max_ttl and num_uses are respected, the rest is for testing purposes
        return {
            "explicit_max_ttl": 1337,
            "num_uses": 42,
            "secret_id_num_uses": 3,
            "secret_id_ttl": 1338,
            "irrelevant_setting": "abc",
        }

    @pytest.mark.usefixtures("conn_cache_absent")
    @pytest.mark.parametrize(
        "vault_container_version", ["0.9.6", "1.3.1", "latest"], indirect=True
    )
    def test_minion_can_authenticate(self, vault_salt_call_cli):
        """
        Test that the minion can run queries against Vault.
        The master impersonating the minion is already tested in the fixture setup
        (ext_pillar).
        """
        ret = vault_salt_call_cli.run("vault.read_secret", "secret/path/foo")
        assert ret.returncode == 0
        assert ret.data
        assert ret.data.get("success") == "yeehaaw"

    @pytest.mark.usefixtures("conn_cache_absent")
    @pytest.mark.parametrize(
        "vault_container_version", ["0.9.6", "1.3.1", "latest"], indirect=True
    )
    def test_minion_token_policies_are_assigned_as_expected(
        self, vault_salt_call_cli, vault_salt_minion
    ):
        """
        Test that issued tokens have the expected policies.
        """
        ret = vault_salt_call_cli.run("vault.query", "GET", "auth/token/lookup-self")
        assert ret.returncode == 0
        assert ret.data
        assert set(ret.data["data"]["policies"]) == {
            "default",
            "salt_minion",
            f"salt_minion_{vault_salt_minion.id}",
            "salt_role_dev",
            "salt_role_web",
        }

    @pytest.mark.parametrize("vault_container_version", ["latest"], indirect=True)
    @pytest.mark.usefixtures("cache_auth_outdated")
    def test_auth_method_switch_does_not_break_minion_auth(
        self, vault_salt_call_cli, caplog
    ):
        """
        Test that after a master configuration switch from another authentication method,
        minions with cached configuration flush it and request a new one.
        """
        ret = vault_salt_call_cli.run("vault.read_secret", "secret/path/foo")
        assert ret.returncode == 0
        assert ret.data
        assert ret.data.get("success") == "yeehaaw"
        assert "Master returned error and requested cache expiration" in caplog.text

    @pytest.mark.parametrize("vault_container_version", ["latest"], indirect=True)
    @pytest.mark.parametrize("ckey", ["config", "__token"])
    def test_cache_is_used_on_the_minion(
        self, ckey, vault_salt_call_cli, minion_conn_cachedir
    ):
        """
        Test that remote configuration and tokens are written to cache.
        """
        cache = minion_conn_cachedir
        if ckey == "__token":
            cache = cache / "session"
            if not cache.exists():
                cache.mkdir()
        if f"{ckey}.p" not in os.listdir(cache):
            ret = vault_salt_call_cli.run("vault.read_secret", "secret/path/foo")
            assert ret.returncode == 0
        assert f"{ckey}.p" in os.listdir(cache)

    @pytest.mark.parametrize("vault_container_version", ["latest"], indirect=True)
    @pytest.mark.parametrize("ckey", ["config", "__token"])
    def test_cache_is_used_on_the_impersonating_master(
        self, ckey, vault_salt_run_cli, vault_salt_minion
    ):
        """
        Test that remote configuration and tokens are written to cache when a
        master is impersonating a minion during pillar rendering.
        """
        cbank = f"minions/{vault_salt_minion.id}/vault/connection"
        if ckey == "__token":
            cbank += "/session"
        ret = vault_salt_run_cli.run("cache.list", cbank)
        assert ret.returncode == 0
        assert ret.data
        assert ckey in ret.data

    @pytest.mark.usefixtures("conn_cache_absent")
    @pytest.mark.parametrize("vault_container_version", ["latest"], indirect=True)
    def test_issue_param_overrides_require_setting(self, overriding_vault_salt_minion):
        """
        Test that minion overrides of issue params are not set by default
        and require setting ``issue:allow_minion_override_params``.
        """
        ret = overriding_vault_salt_minion.salt_call_cli().run(
            "vault.query", "GET", "auth/token/lookup-self"
        )
        assert ret.returncode == 0
        assert ret.data
        assert ret.data["data"]["explicit_max_ttl"] != 1337
        assert ret.data["data"]["num_uses"] != 41  # one use is consumed by the lookup


@pytest.mark.usefixtures("vault_testing_values")
@pytest.mark.parametrize("vault_container_version", ["latest"], indirect=True)
class TestAppRoleIssuanceWithoutSecretId:
    @pytest.fixture(scope="class")
    def vault_master_config(self, vault_port):
        return {
            "open_mode": True,
            "peer_run": {
                ".*": [
                    "vault.get_config",
                    "vault.generate_secret_id",
                ],
            },
            "vault": {
                "auth": {"token": "testsecret"},
                "cache": {
                    "backend": "file",
                },
                "issue": {
                    "type": "approle",
                    "approle": {
                        "params": {
                            "bind_secret_id": False,
                            # "at least one constraint should be enabled on the role"
                            # this should be quite secure :)
                            "token_bound_cidrs": "0.0.0.0/0",
                            "token_explicit_max_ttl": 1800,
                            "token_num_uses": 0,
                        }
                    },
                },
                "policies": {
                    "assign": {
                        "salt_minion",
                        "salt_minion_{minion}",
                    },
                },
                "server": {
                    "url": f"http://127.0.0.1:{vault_port}",
                },
            },
        }

    @pytest.mark.usefixtures("conn_cache_absent")
    def test_minion_can_authenticate(self, vault_salt_call_cli, caplog):
        """
        Test that the minion can run queries against Vault.
        The master impersonating the minion is already tested in the fixture setup
        (ext_pillar).
        """
        ret = vault_salt_call_cli.run("vault.read_secret", "secret/path/foo")
        assert ret.returncode == 0
        assert ret.data
        assert ret.data.get("success") == "yeehaaw"
        assert "Minion AppRole does not require a secret ID" not in caplog.text


@pytest.mark.usefixtures("vault_testing_values")
@pytest.mark.parametrize("vault_container_version", ["latest"], indirect=True)
class TestOldConfigSyntax:
    @pytest.fixture(scope="class")
    def vault_master_config(self, pillar_state_tree, vault_port):
        return {
            "pillar_roots": {"base": [str(pillar_state_tree)]},
            "open_mode": True,
            "peer_run": {
                ".*": [
                    "vault.generate_token",
                ],
            },
            "vault": {
                "auth": {
                    "allow_minion_override": True,
                    "token": "testsecret",
                    "token_backend": "file",
                    "ttl": 90,
                    "uses": 3,
                },
                "policies": [
                    "salt_minion",
                    "salt_minion_{minion}",
                ],
                "url": f"http://127.0.0.1:{vault_port}",
            },
            "minion_data_cache": True,
        }

    @pytest.fixture(scope="class")
    def overriding_vault_salt_minion(self, vault_salt_master):
        assert vault_salt_master.is_running()
        factory = vault_salt_master.salt_minion_daemon(
            random_string("vault-minion", uppercase=False),
            defaults={"open_mode": True, "grains": {}},
            overrides={"vault": {"auth": {"uses": 5, "ttl": 180}}},
        )
        with factory.started():
            # Sync All
            salt_call_cli = factory.salt_call_cli()
            ret = salt_call_cli.run("saltutil.sync_all", _timeout=120)
            assert ret.returncode == 0, ret
            yield factory

    @pytest.mark.usefixtures("conn_cache_absent")
    def test_minion_can_authenticate(self, vault_salt_call_cli, caplog):
        """
        Test that the minion can authenticate, even if the master peer_run
        configuration has not been updated.
        """
        ret = vault_salt_call_cli.run("vault.read_secret", "secret/path/foo")
        assert ret.returncode == 0
        assert ret.data
        assert ret.data.get("success") == "yeehaaw"
        assert (
            "does the peer runner publish configuration include `vault.get_config`"
            in caplog.text
        )
        assert "Peer runner return was empty." not in caplog.text
        assert "Falling back to vault.generate_token." in caplog.text
        assert (
            "Detected minion fallback to old vault.generate_token peer run function"
            in caplog.text
        )

    @pytest.mark.usefixtures("conn_cache_absent")
    def test_token_is_configured_as_expected(
        self, vault_salt_call_cli, vault_salt_minion
    ):
        """
        Test that issued tokens have the expected parameters.
        """
        ret = vault_salt_call_cli.run("vault.query", "GET", "auth/token/lookup-self")
        assert ret.returncode == 0
        assert ret.data
        assert ret.data["data"]["explicit_max_ttl"] == 90
        assert ret.data["data"]["num_uses"] == 2  # one use is consumed by the lookup
        assert set(ret.data["data"]["policies"]) == {
            "default",
            "salt_minion",
            f"salt_minion_{vault_salt_minion.id}",
        }

    @pytest.mark.usefixtures("conn_cache_absent")
    def test_issue_param_overrides_work(self, overriding_vault_salt_minion):
        """
        Test that minion overrides of issue params work for the old configuration.
        """
        ret = overriding_vault_salt_minion.salt_call_cli().run(
            "vault.query", "GET", "auth/token/lookup-self"
        )
        assert ret.returncode == 0
        assert ret.data
        assert ret.data["data"]["explicit_max_ttl"] == 180
        assert ret.data["data"]["num_uses"] == 4  # one use is consumed by the lookup


@pytest.mark.usefixtures("vault_testing_values")
class TestMinionLocal:
    @pytest.fixture(scope="class")
    def vault_master_config(self):
        return {"open_mode": True}

    @pytest.fixture(scope="class")
    def vault_salt_minion(self, vault_salt_master, vault_port):
        assert vault_salt_master.is_running()
        factory = vault_salt_master.salt_minion_daemon(
            random_string("vault-minion", uppercase=False),
            defaults={
                "open_mode": True,
                "vault": {
                    "auth": {"token": "testsecret"},
                    "cache": {
                        "backend": "file",
                    },
                    "server": {
                        "url": f"http://127.0.0.1:{vault_port}",
                    },
                },
                "grains": {},
            },
        )
        with factory.started():
            # Sync All
            salt_call_cli = factory.salt_call_cli()
            ret = salt_call_cli.run("saltutil.sync_all", _timeout=120)
            assert ret.returncode == 0, ret
            yield factory

    def test_minion_can_authenticate(self, vault_salt_call_cli):
        """
        Test that salt-call --local works with the Vault module.
        Issue #58580
        """
        ret = vault_salt_call_cli.run("--local", "vault.read_secret", "secret/path/foo")
        assert ret.returncode == 0
        assert ret.data
        assert ret.data.get("success") == "yeehaaw"



import salt.modules.cmdmod
import salt.utils.json
import salt.utils.platform
from salt.exceptions import CommandExecutionError


def run_dict(cmd, cwd=None):
    """
    Execute the powershell command and return the data as a dictionary

    .. versionadded:: 3006.9

    Args:

        cmd (str,list): The powershell command to run

        cwd (str): The current working directory

    Returns:
        dict: A dictionary containing the output of the powershell command

    Raises:
        CommandExecutionError:
            If an error is encountered or the command does not complete
            successfully
    """
    if isinstance(cmd, list):
        cmd = " ".join(map(str, cmd))
    if "convertto-json" not in cmd.lower():
        cmd = f"{cmd} | ConvertTo-Json"
    if "progresspreference" not in cmd.lower():
        cmd = f"$ProgressPreference = 'SilentlyContinue'; {cmd}"
    ret = salt.modules.cmdmod.run_all(cmd=cmd, shell="powershell", cwd=cwd)

    if "pid" in ret:
        del ret["pid"]

    if ret.get("stderr", ""):
        error = ret["stderr"].splitlines()[0]
        raise CommandExecutionError(error, info=ret)

    if "retcode" not in ret or ret["retcode"] != 0:
        # run_all logs an error to log.error, fail hard back to the user
        raise CommandExecutionError("Issue executing PowerShell cmd", info=ret)

    # Sometimes Powershell returns an empty string, which isn't valid JSON
    if ret["stdout"] == "":
        ret["stdout"] = "{}"

    try:
        ret = salt.utils.json.loads(ret["stdout"], strict=False)
    except ValueError:
        raise CommandExecutionError("No JSON results from PowerShell", info=ret)

    return ret



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""

import logging
import os
import shutil
import textwrap

import pytest

import salt.modules.mount as mount
import salt.utils.files
import salt.utils.path
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, mock_open, patch

log = logging.getLogger(__name__)


@pytest.fixture
def mock_shell_file():
    return "A B C D F G\n"


@pytest.fixture
def config_initial_file():
    inital_fsystem = [
        "/:\n",
        "\tdev\t\t= /dev/hd4\n",
        "\tvfs\t\t= jfs2\n",
        "\tlog\t\t= /dev/hd8\n",
        "\tmount \t\t= automatic\n",
        "\tcheck\t\t= false\n",
        "\ttype\t\t= bootfs\n",
        "\tvol\t\t= root\n",
        "\tfree\t\t= true\n",
        "\n",
        "/home:\n",
        "\tdev\t\t= /dev/hd1\n",
        "\tvfs\t\t= jfs2\n",
        "\tlog\t\t= /dev/hd8\n",
        "\tmount\t\t= true\n",
        "\tcheck\t\t= true\n",
        "\tvol\t\t= /home\n",
        "\tfree\t\t= false\n",
        "\n",
    ]
    return inital_fsystem


@pytest.fixture
def configure_loader_modules():
    return {mount: {}}


@pytest.fixture
def tmp_sub_dir(tmp_path):
    directory = tmp_path / "filesystems-dir"
    directory.mkdir()

    yield directory

    shutil.rmtree(str(directory))


@pytest.fixture
def config_file(tmp_sub_dir, config_initial_file):
    filename = str(tmp_sub_dir / "filesystems")

    with salt.utils.files.fopen(filename, "wb") as fp:
        fp.writelines(salt.utils.data.encode(config_initial_file))

    yield filename

    os.remove(filename)


def test_active():
    """
    List the active mounts.
    """
    with patch.dict(mount.__grains__, {"os": "FreeBSD", "kernel": "FreeBSD"}):
        # uid=user1 tests the improbable case where a OS returns a name
        # instead of a numeric id, for #25293
        mock = MagicMock(return_value="A B C D,E,F,uid=user1,gid=grp1")
        mock_user = MagicMock(return_value={"uid": "100"})
        mock_group = MagicMock(return_value={"gid": "100"})
        with patch.dict(
            mount.__salt__,
            {
                "cmd.run_stdout": mock,
                "user.info": mock_user,
                "group.info": mock_group,
            },
        ):
            assert mount.active() == {
                "B": {
                    "device": "A",
                    "opts": ["D", "E", "F", "uid=100", "gid=100"],
                    "fstype": "C",
                }
            }

    with patch.dict(mount.__grains__, {"os": "Solaris", "kernel": "SunOS"}):
        mock = MagicMock(return_value="A * B * C D/E/F")
        with patch.dict(mount.__salt__, {"cmd.run_stdout": mock}):
            assert mount.active() == {
                "B": {"device": "A", "opts": ["D", "E", "F"], "fstype": "C"}
            }

    with patch.dict(mount.__grains__, {"os": "AIX", "kernel": "AIX"}):
        mock = MagicMock(return_value="A * B * C D/E/F")
        with patch.dict(mount.__salt__, {"cmd.run_stdout": mock}):
            assert mount.active() == {"B": {"node": "A", "device": "*", "fstype": "*"}}

    with patch.dict(mount.__grains__, {"os": "OpenBSD", "kernel": "OpenBSD"}):
        mock = MagicMock(return_value={})
        with patch.object(mount, "_active_mounts_openbsd", mock):
            assert mount.active() == {}

    with patch.dict(mount.__grains__, {"os": "MacOS", "kernel": "Darwin"}):
        mock = MagicMock(return_value={})
        with patch.object(mount, "_active_mounts_darwin", mock):
            assert mount.active() == {}

    with patch.dict(mount.__grains__, {"os": "MacOS", "kernel": "Darwin"}):
        mock = MagicMock(return_value={})
        with patch.object(mount, "_active_mountinfo", mock):
            with patch.object(mount, "_active_mounts_darwin", mock):
                assert mount.active(extended=True) == {}

    with patch.dict(mount.__grains__, {"os": "AIX", "kernel": "AIX"}):
        mock = MagicMock(return_value={})
        with patch.object(mount, "_active_mounts_aix", mock):
            assert mount.active() == {}


def test_fstab_entry_ignores_opt_ordering():
    entry = mount._fstab_entry(
        name="/tmp",
        device="tmpfs",
        fstype="tmpfs",
        opts="defaults,nodev,noexec",
        dump=0,
        pass_num=0,
    )
    assert entry.match("tmpfs\t\t/tmp\ttmpfs\tnodev,defaults,noexec\t0 0\n")


def test_fstab():
    """
    List the content of the fstab
    """
    mock = MagicMock(return_value=False)
    with patch.object(os.path, "isfile", mock):
        assert mount.fstab() == {}

    file_data = "\n".join(["#", "A B C D,E,F G H"])
    mock = MagicMock(return_value=True)
    with patch.dict(mount.__grains__, {"kernel": ""}), patch.object(
        os.path, "isfile", mock
    ), patch("salt.utils.files.fopen", mock_open(read_data=file_data)):
        fstab = mount.fstab()
        assert fstab == {
            "B": {
                "device": "A",
                "dump": "G",
                "fstype": "C",
                "opts": ["D", "E", "F"],
                "pass": "H",
            }
        }, fstab


def test_vfstab():
    """
    List the content of the vfstab
    """
    mock = MagicMock(return_value=False)
    with patch.object(os.path, "isfile", mock):
        assert mount.vfstab() == {}

    file_data = textwrap.dedent(
        """\
        #
        swap        -   /tmp                tmpfs    -   yes    size=2048m
        """
    )
    mock = MagicMock(return_value=True)
    with patch.dict(mount.__grains__, {"kernel": "SunOS"}), patch.object(
        os.path, "isfile", mock
    ), patch("salt.utils.files.fopen", mock_open(read_data=file_data)):
        vfstab = mount.vfstab()
        assert vfstab == {
            "/tmp": {
                "device": "swap",
                "device_fsck": "-",
                "fstype": "tmpfs",
                "mount_at_boot": "yes",
                "opts": ["size=2048m"],
                "pass_fsck": "-",
            }
        }, vfstab


def test_filesystems():
    """
    List the content of the filesystems
    """
    file_data = textwrap.dedent(
        """\
        #

        """
    )
    mock = MagicMock(return_value=True)
    with patch.dict(mount.__grains__, {"os": "AIX", "kernel": "AIX"}), patch.object(
        os.path, "isfile", mock
    ), patch("salt.utils.files.fopen", mock_open(read_data=file_data)):
        assert mount.filesystems() == {}

    file_data = textwrap.dedent(
        """\
        #
        /home:
                dev             = /dev/hd1
                vfs             = jfs2
                log             = /dev/hd8
                mount           = true
                check           = true
                vol             = /home
                free            = false
                quota           = no

        """
    )
    mock = MagicMock(return_value=True)
    with patch.dict(mount.__grains__, {"os": "AIX", "kernel": "AIX"}), patch.object(
        os.path, "isfile", mock
    ), patch("salt.utils.files.fopen", mock_open(read_data=file_data)):
        fsyst = mount.filesystems()
        test_fsyst = {
            "/home": {
                "dev": "/dev/hd1",
                "vfs": "jfs2",
                "log": "/dev/hd8",
                "mount": "true",
                "check": "true",
                "vol": "/home",
                "free": "false",
                "quota": "no",
            }
        }
        assert test_fsyst == fsyst


def test_rm_fstab():
    """
    Remove the mount point from the fstab
    """
    mock_fstab = MagicMock(return_value={})
    with patch.dict(mount.__grains__, {"kernel": ""}):
        with patch.object(mount, "fstab", mock_fstab):
            with patch("salt.utils.files.fopen", mock_open()):
                assert mount.rm_fstab("name", "device")


def test_set_fstab(mock_shell_file):
    """
    Tests to verify that this mount is represented in the fstab,
    change the mount to match the data passed, or add the mount
    if it is not present.
    """
    mock = MagicMock(return_value=False)
    with patch.object(os.path, "isfile", mock):
        pytest.raises(CommandExecutionError, mount.set_fstab, "A", "B", "C")

    mock = MagicMock(return_value=True)
    mock_read = MagicMock(side_effect=OSError)
    with patch.object(os.path, "isfile", mock):
        with patch.object(salt.utils.files, "fopen", mock_read):
            pytest.raises(CommandExecutionError, mount.set_fstab, "A", "B", "C")

    mock = MagicMock(return_value=True)
    with patch.object(os.path, "isfile", mock):
        with patch("salt.utils.files.fopen", mock_open(read_data=mock_shell_file)):
            assert mount.set_fstab("A", "B", "C") == "new"

    mock = MagicMock(return_value=True)
    with patch.object(os.path, "isfile", mock):
        with patch("salt.utils.files.fopen", mock_open(read_data=mock_shell_file)):
            assert mount.set_fstab("B", "A", "C", "D", "F", "G") == "present"

    mock = MagicMock(return_value=True)
    with patch.object(os.path, "isfile", mock):
        with patch("salt.utils.files.fopen", mock_open(read_data=mock_shell_file)):
            assert mount.set_fstab("B", "A", "C", not_change=True) == "present"


def test_rm_automaster():
    """
    Remove the mount point from the auto_master
    """
    mock = MagicMock(return_value={})
    with patch.object(mount, "automaster", mock):
        assert mount.rm_automaster("name", "device")

    mock = MagicMock(return_value={"name": "name"})
    with patch.object(mount, "fstab", mock):
        assert mount.rm_automaster("name", "device")


def test_set_automaster(mock_shell_file):
    """
    Verify that this mount is represented in the auto_salt, change the mount
    to match the data passed, or add the mount if it is not present.
    """
    mock = MagicMock(return_value=True)
    with patch.object(os.path, "isfile", mock):
        pytest.raises(CommandExecutionError, mount.set_automaster, "A", "B", "C")

    mock = MagicMock(return_value=True)
    mock_read = MagicMock(side_effect=OSError)
    with patch.object(os.path, "isfile", mock):
        with patch.object(salt.utils.files, "fopen", mock_read):
            pytest.raises(CommandExecutionError, mount.set_automaster, "A", "B", "C")

    mock = MagicMock(return_value=True)
    with patch.object(os.path, "isfile", mock):
        with patch("salt.utils.files.fopen", mock_open(read_data=mock_shell_file)):
            assert mount.set_automaster("A", "B", "C") == "new"

    mock = MagicMock(return_value=True)
    with patch.object(os.path, "isfile", mock):
        with patch(
            "salt.utils.files.fopen", mock_open(read_data="/..A -fstype=C,D C:B")
        ):
            assert mount.set_automaster("A", "B", "C", "D") == "present"

    mock = MagicMock(return_value=True)
    with patch.object(os.path, "isfile", mock):
        with patch(
            "salt.utils.files.fopen", mock_open(read_data="/..A -fstype=XX C:B")
        ):
            assert (
                mount.set_automaster("A", "B", "C", "D", not_change=True) == "present"
            )


def test_automaster():
    """
    Test the list the contents of the fstab
    """
    assert mount.automaster() == {}


def test_rm_filesystems():
    """
    Remove the mount point from the filesystems
    """
    file_data = textwrap.dedent(
        """\
        #

        """
    )
    mock = MagicMock(return_value=True)
    with patch.dict(mount.__grains__, {"os": "AIX", "kernel": "AIX"}), patch.object(
        os.path, "isfile", mock
    ), patch("salt.utils.files.fopen", mock_open(read_data=file_data)):
        assert not mount.rm_filesystems("name", "device")

    file_data = textwrap.dedent(
        """\
        #
        /name:
                dev             = device
                vol             = /name

        """
    )

    mock = MagicMock(return_value=True)
    mock_fsyst = MagicMock(return_value=True)
    with patch.dict(mount.__grains__, {"os": "AIX", "kernel": "AIX"}), patch.object(
        os.path, "isfile", mock
    ), patch("salt.utils.files.fopen", mock_open(read_data=file_data)):
        assert mount.rm_filesystems("/name", "device")


def test_set_filesystems():
    """
    Tests to verify that this mount is represented in the filesystems,
    change the mount to match the data passed, or add the mount
    if it is not present.
    """
    mock = MagicMock(return_value=False)
    with patch.dict(mount.__grains__, {"os": "AIX", "kernel": "AIX"}):
        with patch.object(os.path, "isfile", mock):
            pytest.raises(CommandExecutionError, mount.set_filesystems, "A", "B", "C")

        mock_read = MagicMock(side_effect=OSError)
        with patch.object(os.path, "isfile", mock):
            with patch.object(salt.utils.files, "fopen", mock_read):
                pytest.raises(
                    CommandExecutionError, mount.set_filesystems, "A", "B", "C"
                )


@pytest.mark.skip_on_windows(
    reason="Not supported on Windows, does not handle tabs well"
)
def test_set_filesystems_with_data(tmp_sub_dir, config_file):
    """
    Tests to verify set_filesystems reads and adjusts file /etc/filesystems correctly
    """
    # Note AIX uses tabs in filesystems files, hence disable warings and errors for tabs and spaces
    config_filepath = str(tmp_sub_dir / "filesystems")
    with patch.dict(mount.__grains__, {"os": "AIX", "kernel": "AIX"}):
        mount.set_filesystems(
            "/test_mount", "/dev/hd3", "jsf2", "-", "true", config_filepath
        )
        with salt.utils.files.fopen(config_filepath, "r") as fp:
            fsys_content = fp.read()

        test_fsyst = """/:
	dev		= /dev/hd4
	vfs		= jfs2
	log		= /dev/hd8
	mount		= automatic
	check		= false
	type		= bootfs
	vol		= root
	free		= true

/home:
	dev		= /dev/hd1
	vfs		= jfs2
	log		= /dev/hd8
	mount		= true
	check		= true
	vol		= /home
	free		= false

/test_mount:
	dev		= /dev/hd3
	vfstype		= jsf2
	opts		= -
	mount		= true

"""
    assert test_fsyst == fsys_content


def test_mount():
    """
    Mount a device
    """
    with patch.dict(mount.__grains__, {"os": "MacOS"}):
        mock = MagicMock(return_value=True)
        with patch.object(os.path, "exists", mock):
            mock = MagicMock(return_value=None)
            with patch.dict(mount.__salt__, {"file.mkdir": None}):
                mock = MagicMock(return_value={"retcode": True, "stderr": True})
                with patch.dict(mount.__salt__, {"cmd.run_all": mock}):
                    assert mount.mount("name", "device")
                    mock.assert_called_with(
                        "mount  'device' 'name' ", python_shell=False, runas=None
                    )

                with patch.dict(mount.__salt__, {"cmd.run_all": mock}):
                    assert mount.mount("name", "device", fstype="fstype")
                    mock.assert_called_with(
                        "mount  -t fstype 'device' 'name' ",
                        python_shell=False,
                        runas=None,
                    )

                mock = MagicMock(return_value={"retcode": False, "stderr": False})
                with patch.dict(mount.__salt__, {"cmd.run_all": mock}):
                    assert mount.mount("name", "device")

    with patch.dict(mount.__grains__, {"os": "AIX"}):
        mock = MagicMock(return_value=True)
        with patch.object(os.path, "exists", mock):
            mock = MagicMock(return_value=None)
            with patch.dict(mount.__salt__, {"file.mkdir": None}):
                mock = MagicMock(return_value={"retcode": True, "stderr": True})
                with patch.dict(mount.__salt__, {"cmd.run_all": mock}):
                    assert mount.mount("name", "device")
                    mock.assert_called_with(
                        "mount  'device' 'name' ", python_shell=False, runas=None
                    )

                with patch.dict(mount.__salt__, {"cmd.run_all": mock}):
                    assert mount.mount("name", "device", fstype="fstype")
                    mock.assert_called_with(
                        "mount  -v fstype 'device' 'name' ",
                        python_shell=False,
                        runas=None,
                    )

                mock = MagicMock(return_value={"retcode": False, "stderr": False})
                with patch.dict(mount.__salt__, {"cmd.run_all": mock}):
                    assert mount.mount("name", "device")

    with patch.dict(mount.__grains__, {"os": "Linux"}):
        mock = MagicMock(return_value=True)
        with patch.object(os.path, "exists", mock):
            mock = MagicMock(return_value=None)
            with patch.dict(mount.__salt__, {"file.mkdir": None}):
                mock = MagicMock(return_value={"retcode": True, "stderr": True})
                with patch.dict(mount.__salt__, {"cmd.run_all": mock}):
                    assert mount.mount("name", "device")
                    mock.assert_called_with(
                        "mount -o defaults 'device' 'name' ",
                        python_shell=False,
                        runas=None,
                    )

                with patch.dict(mount.__salt__, {"cmd.run_all": mock}):
                    assert mount.mount("name", "device", fstype="fstype")
                    mock.assert_called_with(
                        "mount -o defaults -t fstype 'device' 'name' ",
                        python_shell=False,
                        runas=None,
                    )

                mock = MagicMock(return_value={"retcode": False, "stderr": False})
                with patch.dict(mount.__salt__, {"cmd.run_all": mock}):
                    assert mount.mount("name", "device")


def test_remount_non_mounted():
    """
    Attempt to remount a device, if the device is not already mounted, mount
    is called
    """
    with patch.dict(mount.__grains__, {"os": "MacOS"}):
        mock = MagicMock(return_value=[])
        with patch.object(mount, "active", mock):
            mock = MagicMock(return_value=True)
            with patch.object(mount, "mount", mock):
                assert mount.remount("name", "device")

    with patch.dict(mount.__grains__, {"os": "AIX"}):
        mock = MagicMock(return_value=[])
        with patch.object(mount, "active", mock):
            mock = MagicMock(return_value=True)
            with patch.object(mount, "mount", mock):
                assert mount.remount("name", "device")

    with patch.dict(mount.__grains__, {"os": "Linux"}):
        mock = MagicMock(return_value=[])
        with patch.object(mount, "active", mock):
            mock = MagicMock(return_value=True)
            with patch.object(mount, "mount", mock):
                assert mount.remount("name", "device")


def test_remount_already_mounted_no_fstype():
    """
    Attempt to remount a device already mounted that do not provides
    fstype
    """
    with patch.dict(mount.__grains__, {"os": "MacOS"}):
        mock = MagicMock(return_value=["name"])
        with patch.object(mount, "active", mock):
            mock = MagicMock(return_value={"retcode": 0})
            with patch.dict(mount.__salt__, {"cmd.run_all": mock}):
                assert mount.remount("name", "device")
                mock.assert_called_with(
                    "mount -u -o noowners 'device' 'name' ",
                    python_shell=False,
                    runas=None,
                )

    with patch.dict(mount.__grains__, {"os": "AIX"}):
        mock = MagicMock(return_value=["name"])
        with patch.object(mount, "active", mock):
            mock = MagicMock(return_value={"retcode": 0})
            with patch.dict(mount.__salt__, {"cmd.run_all": mock}):
                assert mount.remount("name", "device")
                mock.assert_called_with(
                    "mount -o remount 'device' 'name' ", python_shell=False, runas=None
                )

    with patch.dict(mount.__grains__, {"os": "Linux"}):
        mock = MagicMock(return_value=["name"])
        with patch.object(mount, "active", mock):
            mock = MagicMock(return_value={"retcode": 0})
            with patch.dict(mount.__salt__, {"cmd.run_all": mock}):
                assert mount.remount("name", "device")
                mock.assert_called_with(
                    "mount -o defaults,remount 'device' 'name' ",
                    python_shell=False,
                    runas=None,
                )


def test_remount_already_mounted_with_fstype():
    """
    Attempt to remount a device already mounted that do not provides
    fstype
    """
    with patch.dict(mount.__grains__, {"os": "MacOS"}):
        mock = MagicMock(return_value=["name"])
        with patch.object(mount, "active", mock):
            mock = MagicMock(return_value={"retcode": 0})
            with patch.dict(mount.__salt__, {"cmd.run_all": mock}):
                assert mount.remount("name", "device", fstype="type")
                mock.assert_called_with(
                    "mount -u -o noowners -t type 'device' 'name' ",
                    python_shell=False,
                    runas=None,
                )

    with patch.dict(mount.__grains__, {"os": "AIX"}):
        mock = MagicMock(return_value=["name"])
        with patch.object(mount, "active", mock):
            mock = MagicMock(return_value={"retcode": 0})
            with patch.dict(mount.__salt__, {"cmd.run_all": mock}):
                assert mount.remount("name", "device", fstype="type")
                mock.assert_called_with(
                    "mount -o remount -v type 'device' 'name' ",
                    python_shell=False,
                    runas=None,
                )

    with patch.dict(mount.__grains__, {"os": "Linux"}):
        mock = MagicMock(return_value=["name"])
        with patch.object(mount, "active", mock):
            mock = MagicMock(return_value={"retcode": 0})
            with patch.dict(mount.__salt__, {"cmd.run_all": mock}):
                assert mount.remount("name", "device", fstype="type")
                mock.assert_called_with(
                    "mount -o defaults,remount -t type 'device' 'name' ",
                    python_shell=False,
                    runas=None,
                )


def test_umount():
    """
    Attempt to unmount a device by specifying the directory it is
    mounted on
    """
    mock = MagicMock(return_value={})
    with patch.object(mount, "active", mock):
        assert mount.umount("name") == "name does not have anything mounted"

    mock = MagicMock(return_value={"name": "name"})
    with patch.object(mount, "active", mock):
        mock = MagicMock(return_value={"retcode": True, "stderr": True})
        with patch.dict(mount.__salt__, {"cmd.run_all": mock}):
            assert mount.umount("name")

        mock = MagicMock(return_value={"retcode": False})
        with patch.dict(mount.__salt__, {"cmd.run_all": mock}):
            assert mount.umount("name")

    # Test unmounting with guestfs util
    mock = MagicMock()
    with patch.dict(mount.__salt__, {"guestfs.umount": mock}):
        mount.umount("/mountpoint", device="/path/to/my.qcow", util="guestfs")
        mock.assert_called_once_with("/mountpoint", disk="/path/to/my.qcow")


def test_umount_lazy_true():
    """
    Attempt to lazy unmount a device by specifying the
    directory it is mounted on
    """
    mock_mount_active = MagicMock(return_value={"name": "name"})
    with patch.object(mount, "active", mock_mount_active):
        mock_cmd = MagicMock(return_value={"retcode": True, "stderr": True})
        with patch.dict(mount.__salt__, {"cmd.run_all": mock_cmd}):
            mount.umount("name", lazy=True)
            mock_cmd.assert_called_once_with(
                "umount -l 'name'", runas=None, python_shell=False
            )


def test_is_fuse_exec():
    """
    Returns true if the command passed is a fuse mountable application
    """
    with patch.object(salt.utils.path, "which", return_value=None):
        assert not mount.is_fuse_exec("cmd")

    which_mock = MagicMock(side_effect=lambda x: x)
    ldd_mock = MagicMock(
        side_effect=[
            textwrap.dedent(
                """\
                linux-vdso.so.1 (0x00007ffeaf5fb000)
                libfuse3.so.3 => /usr/lib/libfuse3.so.3 (0x00007f91e66ac000)
                """
            ),
            textwrap.dedent(
                """\
                linux-vdso.so.1 (0x00007ffeaf5fb000)
                """
            ),
        ]
    )
    with patch.object(salt.utils.path, "which", which_mock):
        with patch.dict(mount.__salt__, {"cmd.run": ldd_mock}):
            assert mount.is_fuse_exec("cmd1")
            assert not mount.is_fuse_exec("cmd2")


def test_swaps():
    """
    Return a dict containing information on active swap
    """
    file_data = textwrap.dedent(
        """\
        Filename Type Size Used Priority
        /dev/sda1 partition 31249404 4100 -1
        """
    )
    with patch.dict(mount.__grains__, {"os": "", "kernel": ""}):
        with patch("salt.utils.files.fopen", mock_open(read_data=file_data)):
            swaps = mount.swaps()
            assert swaps == {
                "/dev/sda1": {
                    "priority": "-1",
                    "size": "31249404",
                    "type": "partition",
                    "used": "4100",
                }
            }, swaps

    file_data = textwrap.dedent(
        """\
        Device Size Used Unknown Unknown Priority
        /dev/sda1 31249404 4100 unknown unknown -1
        """
    )
    mock = MagicMock(return_value=file_data)
    with patch.dict(
        mount.__grains__, {"os": "OpenBSD", "kernel": "OpenBSD"}
    ), patch.dict(mount.__salt__, {"cmd.run_stdout": mock}):
        swaps = mount.swaps()
        assert swaps == {
            "/dev/sda1": {
                "priority": "-1",
                "size": "31249404",
                "type": "partition",
                "used": "4100",
            }
        }, swaps

    file_data = textwrap.dedent(
        """\
        device              maj,min        total       free
        /dev/hd6              10,  2     11776MB     11765MB
        """
    )
    mock = MagicMock(return_value=file_data)
    with patch.dict(mount.__grains__, {"os": "AIX", "kernel": "AIX"}), patch.dict(
        mount.__salt__, {"cmd.run_stdout": mock}
    ):
        swaps = mount.swaps()
        assert swaps == {
            "/dev/hd6": {
                "priority": "-",
                "size": 12058624,
                "type": "device",
                "used": 11264,
            }
        }, swaps


def test_swapon():
    """
    Activate a swap disk
    """
    mock = MagicMock(return_value={"name": "name"})
    with patch.dict(mount.__grains__, {"kernel": ""}):
        with patch.object(mount, "swaps", mock):
            assert mount.swapon("name") == {"stats": "name", "new": False}

    mock = MagicMock(return_value={})
    with patch.dict(mount.__grains__, {"kernel": ""}):
        with patch.object(mount, "swaps", mock):
            mock = MagicMock(return_value=None)
            with patch.dict(mount.__salt__, {"cmd.run": mock}):
                assert mount.swapon("name", False) == {}

    mock = MagicMock(side_effect=[{}, {"name": "name"}])
    with patch.dict(mount.__grains__, {"kernel": ""}):
        with patch.object(mount, "swaps", mock):
            mock = MagicMock(return_value=None)
            with patch.dict(mount.__salt__, {"cmd.run": mock}):
                assert mount.swapon("name") == {"stats": "name", "new": True}
    ## effects of AIX
    mock = MagicMock(return_value={"name": "name"})
    with patch.dict(mount.__grains__, {"kernel": "AIX"}):
        with patch.object(mount, "swaps", mock):
            assert mount.swapon("name") == {"stats": "name", "new": False}

    mock = MagicMock(return_value={})
    with patch.dict(mount.__grains__, {"kernel": "AIX"}):
        with patch.object(mount, "swaps", mock):
            mock = MagicMock(return_value=None)
            with patch.dict(mount.__salt__, {"cmd.run": mock}):
                assert mount.swapon("name", False) == {}

    mock = MagicMock(side_effect=[{}, {"name": "name"}])
    with patch.dict(mount.__grains__, {"kernel": "AIX"}):
        with patch.object(mount, "swaps", mock):
            mock = MagicMock(return_value=None)
            with patch.dict(mount.__salt__, {"cmd.run": mock}):
                assert mount.swapon("name") == {"stats": "name", "new": True}


def test_swapoff():
    """
    Deactivate a named swap mount
    """
    mock = MagicMock(return_value={})
    with patch.dict(mount.__grains__, {"kernel": ""}):
        with patch.object(mount, "swaps", mock):
            assert mount.swapoff("name") is None

    mock = MagicMock(return_value={"name": "name"})
    with patch.dict(mount.__grains__, {"kernel": ""}):
        with patch.object(mount, "swaps", mock):
            with patch.dict(mount.__grains__, {"os": "test"}):
                mock = MagicMock(return_value=None)
                with patch.dict(mount.__salt__, {"cmd.run": mock}):
                    assert not mount.swapoff("name")

    mock = MagicMock(side_effect=[{"name": "name"}, {}])
    with patch.dict(mount.__grains__, {"kernel": ""}):
        with patch.object(mount, "swaps", mock):
            with patch.dict(mount.__grains__, {"os": "test"}):
                mock = MagicMock(return_value=None)
                with patch.dict(mount.__salt__, {"cmd.run": mock}):
                    assert mount.swapoff("name")

    # check on AIX
    mock = MagicMock(return_value={})
    with patch.dict(mount.__grains__, {"kernel": "AIX"}):
        with patch.object(mount, "swaps", mock):
            assert mount.swapoff("name") is None

    mock = MagicMock(return_value={"name": "name"})
    with patch.dict(mount.__grains__, {"kernel": "AIX"}):
        with patch.object(mount, "swaps", mock):
            with patch.dict(mount.__grains__, {"os": "test"}):
                mock = MagicMock(return_value=None)
                with patch.dict(mount.__salt__, {"cmd.run": mock}):
                    assert not mount.swapoff("name")

    mock = MagicMock(side_effect=[{"name": "name"}, {}])
    with patch.dict(mount.__grains__, {"kernel": "AIX"}):
        with patch.object(mount, "swaps", mock):
            with patch.dict(mount.__grains__, {"os": "test"}):
                mock = MagicMock(return_value=None)
                with patch.dict(mount.__salt__, {"cmd.run": mock}):
                    assert mount.swapoff("name")


def test_is_mounted():
    """
    Provide information if the path is mounted
    """
    mock = MagicMock(return_value={})
    with patch.object(mount, "active", mock), patch.dict(
        mount.__grains__, {"kernel": ""}
    ):
        assert not mount.is_mounted("name")

    mock = MagicMock(return_value={"name": "name"})
    with patch.object(mount, "active", mock), patch.dict(
        mount.__grains__, {"kernel": ""}
    ):
        assert mount.is_mounted("name")


def test_get_mount_from_path(tmp_path):
    expected = tmp_path
    while not os.path.ismount(expected):
        expected = expected.parent
    path = str(tmp_path)
    ret = mount.get_mount_from_path(path)
    assert ret == str(expected)


def test_get_device_from_path(tmp_path):
    expected = tmp_path
    while not os.path.ismount(expected):
        expected = expected.parent
    mock_active = [
        {},
        {str(expected): {"device": "mydevice"}},
    ]
    path = str(tmp_path)
    with patch("salt.modules.mount.active", MagicMock(side_effect=mock_active)):
        with patch.dict(mount.__grains__, {"kernel": ""}):
            with patch.dict(mount.__grains__, {"os": "test"}):
                ret = mount.get_device_from_path(path)
                assert ret is None
                ret = mount.get_device_from_path(path)
                assert ret == "mydevice"



import os

import salt.loader
import salt.minion
import salt.utils.yaml
from salt.utils.files import fopen
from tests.support.case import ModuleCase
from tests.support.helpers import with_tempdir
from tests.support.mock import patch


class BasePillarTest(ModuleCase):
    @with_tempdir()
    def test_minion_cache_should_cache_files(self, tempdir):
        pillar = {"this": {"is": {"some": "pillar data"}}}
        opts = {
            "file_client": "remote",
            "minion_pillar_cache": "true",
            "master_type": "local",
            "discovery": False,
            "master": "local",
            "__role": "",
            "id": "test",
            "saltenv": "base",
            "pillar_cache": True,
            "pillar_cache_backend": "disk",
            "pillar_cache_ttl": 3600,
            "cachedir": tempdir,
            "state_top": "top.sls",
            "pillar_roots": {"base": tempdir},
            "extension_modules": tempdir,
            "file_ignore_regex": [],
            "file_ignore_glob": [],
            "pillar": pillar,
        }
        with patch("salt.loader.grains", return_value={}), patch(
            "salt.minion.SMinion.gen_modules"
        ), patch("tornado.ioloop.IOLoop.current"):
            minion = salt.minion.SMinion(opts)
            self.assertTrue("pillar" in os.listdir(tempdir))
            pillar_cache = os.path.join(tempdir, "pillar")
            self.assertTrue("top.sls" in os.listdir(pillar_cache))
            self.assertTrue("cache.sls" in os.listdir(pillar_cache))
            with fopen(os.path.join(pillar_cache, "cache.sls"), "rb") as f:
                cached_data = salt.utils.yaml.safe_load(f)
                assert cached_data == pillar



import asyncio

import salt.transport


async def test_publsh_server(
    io_loop, minion_opts, master_opts, transport, process_manager
):
    minion_opts["transport"] = master_opts["transport"] = transport

    pub_server = salt.transport.publish_server(master_opts)
    pub_server.pre_fork(process_manager)
    await asyncio.sleep(3)

    pub_client = salt.transport.publish_client(
        minion_opts, io_loop, master_opts["interface"], master_opts["publish_port"]
    )
    await pub_client.connect()

    # Yield to loop in order to allow pub client to connect.
    event = asyncio.Event()

    messages = []

    async def handle_msg(msg):
        messages.append(msg)
        event.set()

    try:
        pub_client.on_recv(handle_msg)
        # TODO: Fix this inconsistancy.
        if transport == "zeromq":
            msg = b"meh"
        else:
            msg = {b"foo": b"bar"}
        await pub_server.publish(msg)
        await asyncio.wait_for(event.wait(), 1)
        assert [msg] == messages
    finally:
        pub_server.close()
        pub_client.close()

    # Yield to loop in order to allow background close methods to finish.
    await asyncio.sleep(0.3)



import time

import pytest

import salt.auth
import salt.config
import salt.exceptions
import salt.master
from tests.support.mock import MagicMock, call, patch


@pytest.fixture
def load_auth():
    patches = (
        (
            "salt.loader.auth",
            MagicMock(
                return_value={
                    "pam.auth": "fake_func_str",
                    "pam.groups": "fake_groups_function_str",
                }
            ),
        ),
        (
            "salt.loader.eauth_tokens",
            MagicMock(
                return_value={
                    "localfs.mk_token": "fake_func_mktok",
                    "localfs.get_token": "fake_func_gettok",
                    "localfs.rm_roken": "fake_func_rmtok",
                }
            ),
        ),
    )
    patchers = []
    for mod, mock in patches:
        patcher = patch(mod, mock)
        patcher.start()
        patchers.append(patcher)
    lauth = salt.auth.LoadAuth({})  # Load with empty opts
    try:
        yield lauth
    finally:
        for patcher in patchers:
            patcher.stop()


@pytest.fixture
def master_acl_master_opts(master_opts):
    master_opts["publisher_acl"] = {}
    master_opts["publisher_acl_blacklist"] = {}
    master_opts["master_job_cache"] = ""
    master_opts["sign_pub_messages"] = False
    master_opts["con_cache"] = ""
    master_opts["external_auth"] = {}
    master_opts["external_auth"]["pam"] = {
        "test_user": [
            {"*": ["test.ping"]},
            {"minion_glob*": ["foo.bar"]},
            {"minion_func_test": ["func_test.*"]},
        ],
        "test_group%": [{"*": ["test.echo"]}],
        "test_user_mminion": [{"target_minion": ["test.ping"]}],
        "*": [{"my_minion": ["my_mod.my_func"]}],
        "test_user_func": [
            {
                "*": [
                    {"test.echo": {"args": ["MSG:.*"]}},
                    {
                        "test.echo": {
                            "kwargs": {
                                "text": "KWMSG:.*",
                                "anything": ".*",
                                "none": None,
                            }
                        }
                    },
                    {
                        "my_mod.*": {
                            "args": ["a.*", "b.*"],
                            "kwargs": {"kwa": "kwa.*", "kwb": "kwb"},
                        }
                    },
                ]
            },
            {
                "minion1": [
                    {"test.echo": {"args": ["TEST", None, "TEST.*"]}},
                    {"test.empty": {}},
                ]
            },
        ],
    }
    yield master_opts


@pytest.fixture
def master_acl_clear_funcs(master_acl_master_opts):
    fire_event_mock = MagicMock(return_value="dummy_tag")
    patches = (
        ("zmq.Context", MagicMock()),
        ("salt.payload.dumps", MagicMock()),
        ("salt.master.tagify", MagicMock()),
        ("salt.utils.event.SaltEvent.fire_event", fire_event_mock),
        ("salt.auth.LoadAuth.time_auth", MagicMock(return_value=True)),
        ("salt.minion.MasterMinion", MagicMock()),
        ("salt.utils.verify.check_path_traversal", MagicMock()),
        ("salt.client.get_local_client", MagicMock()),
    )
    patchers = []
    for mod, mock in patches:
        patcher = patch(mod, mock)
        patcher.start()
        patchers.append(patcher)
    clear = salt.master.ClearFuncs(master_acl_master_opts, MagicMock())

    async def _send_pub(*args):
        pass

    clear._send_pub = _send_pub
    try:
        yield clear
    finally:
        for patcher in patchers:
            patcher.stop()


@pytest.fixture
def master_acl_valid_load():
    yield {
        "tgt_type": "glob",
        "jid": "",
        "cmd": "publish",
        "tgt": "test_minion",
        "kwargs": {
            "username": "test_user",
            "password": "test_password",
            "show_timeout": False,
            "eauth": "pam",
            "show_jid": False,
        },
        "ret": "",
        "user": "test_user",
        "key": "",
        "arg": "",
        "fun": "test.ping",
    }


@pytest.fixture
def auth_acl_master_opts(master_opts):
    """
    Master options
    """
    master_opts["publisher_acl"] = {}
    master_opts["publisher_acl_blacklist"] = {}
    master_opts["master_job_cache"] = ""
    master_opts["sign_pub_messages"] = False
    master_opts["con_cache"] = ""
    master_opts["external_auth"] = {}
    master_opts["external_auth"] = {
        "pam": {"test_user": [{"alpha_minion": ["test.ping"]}]}
    }
    yield master_opts


@pytest.fixture
def auth_acl_clear_funcs(auth_acl_master_opts):
    auth_check_mock = MagicMock(return_value=True)
    patches = (
        ("salt.minion.MasterMinion", MagicMock()),
        ("salt.utils.verify.check_path_traversal", MagicMock()),
        ("salt.utils.minions.CkMinions.auth_check", auth_check_mock),
        ("salt.auth.LoadAuth.time_auth", MagicMock(return_value=True)),
        ("salt.client.get_local_client", MagicMock()),
    )
    patchers = []
    for mod, mock in patches:
        patcher = patch(mod, mock)
        patcher.start()
        patchers.append(patcher)
    clear = salt.master.ClearFuncs(auth_acl_master_opts, MagicMock())
    try:
        yield clear
    finally:
        for patcher in patchers:
            patcher.stop()


@pytest.fixture
def auth_acl_valid_load():
    yield {
        "tgt_type": "glob",
        "jid": "",
        "cmd": "publish",
        "tgt": "test_minion",
        "kwargs": {
            "username": "test_user",
            "password": "test_password",
            "show_timeout": False,
            "eauth": "pam",
            "show_jid": False,
        },
        "ret": "",
        "user": "test_user",
        "key": "",
        "arg": "",
        "fun": "test.ping",
    }


def test_get_tok_with_broken_file_will_remove_bad_token(load_auth):
    fake_get_token = MagicMock(
        side_effect=salt.exceptions.SaltDeserializationError("hi")
    )
    patch_opts = patch.dict(load_auth.opts, {"eauth_tokens": "testfs"})
    patch_get_token = patch.dict(
        load_auth.tokens,
        {"testfs.get_token": fake_get_token},
    )
    mock_rm_token = MagicMock()
    patch_rm_token = patch.object(load_auth, "rm_token", mock_rm_token)
    with patch_opts, patch_get_token, patch_rm_token:
        expected_token = "fnord"
        load_auth.get_tok(expected_token)
        mock_rm_token.assert_called_with(expected_token)


def test_get_tok_with_no_expiration_should_remove_bad_token(load_auth):
    fake_get_token = MagicMock(return_value={"no_expire_here": "Nope"})
    patch_opts = patch.dict(load_auth.opts, {"eauth_tokens": "testfs"})
    patch_get_token = patch.dict(
        load_auth.tokens,
        {"testfs.get_token": fake_get_token},
    )
    mock_rm_token = MagicMock()
    patch_rm_token = patch.object(load_auth, "rm_token", mock_rm_token)
    with patch_opts, patch_get_token, patch_rm_token:
        expected_token = "fnord"
        load_auth.get_tok(expected_token)
        mock_rm_token.assert_called_with(expected_token)


def test_get_tok_with_expire_before_current_time_should_remove_token(load_auth):
    fake_get_token = MagicMock(return_value={"expire": time.time() - 1})
    patch_opts = patch.dict(load_auth.opts, {"eauth_tokens": "testfs"})
    patch_get_token = patch.dict(
        load_auth.tokens,
        {"testfs.get_token": fake_get_token},
    )
    mock_rm_token = MagicMock()
    patch_rm_token = patch.object(load_auth, "rm_token", mock_rm_token)
    with patch_opts, patch_get_token, patch_rm_token:
        expected_token = "fnord"
        load_auth.get_tok(expected_token)
        mock_rm_token.assert_called_with(expected_token)


def test_get_tok_with_valid_expiration_should_return_token(load_auth):
    expected_token = {"expire": time.time() + 1}
    fake_get_token = MagicMock(return_value=expected_token)
    patch_opts = patch.dict(load_auth.opts, {"eauth_tokens": "testfs"})
    patch_get_token = patch.dict(
        load_auth.tokens,
        {"testfs.get_token": fake_get_token},
    )
    mock_rm_token = MagicMock()
    patch_rm_token = patch.object(load_auth, "rm_token", mock_rm_token)
    with patch_opts, patch_get_token, patch_rm_token:
        token_name = "fnord"
        actual_token = load_auth.get_tok(token_name)
        mock_rm_token.assert_not_called()
        assert expected_token is actual_token, "Token was not returned"


def test_load_name(load_auth):
    valid_eauth_load = {
        "username": "test_user",
        "show_timeout": False,
        "test_password": "",
        "eauth": "pam",
    }

    # Test a case where the loader auth doesn't have the auth type
    without_auth_type = dict(valid_eauth_load)
    without_auth_type.pop("eauth")
    ret = load_auth.load_name(without_auth_type)
    assert ret == "", "Did not bail when the auth loader didn't have the auth type."

    # Test a case with valid params
    with patch(
        "salt.utils.args.arg_lookup",
        MagicMock(return_value={"args": ["username", "password"]}),
    ) as format_call_mock:
        expected_ret = call("fake_func_str")
        ret = load_auth.load_name(valid_eauth_load)
        format_call_mock.assert_has_calls((expected_ret,), any_order=True)
        assert ret == "test_user"


def test_get_groups(load_auth):
    valid_eauth_load = {
        "username": "test_user",
        "show_timeout": False,
        "test_password": "",
        "eauth": "pam",
    }
    with patch("salt.utils.args.format_call") as format_call_mock:
        expected_ret = call(
            "fake_groups_function_str",
            {
                "username": "test_user",
                "test_password": "",
                "show_timeout": False,
                "eauth": "pam",
            },
            expected_extra_kws=salt.auth.AUTH_INTERNAL_KEYWORDS,
        )
        load_auth.get_groups(valid_eauth_load)
        format_call_mock.assert_has_calls((expected_ret,), any_order=True)


@pytest.mark.skip_on_windows(reason="PAM eauth not available on Windows")
async def test_master_publish_name(master_acl_clear_funcs, master_acl_valid_load):
    """
    Test to ensure a simple name can auth against a given function.
    This tests to ensure test_user can access test.ping but *not* sys.doc
    """
    _check_minions_return = {"minions": ["some_minions"], "missing": []}
    with patch(
        "salt.utils.minions.CkMinions.check_minions",
        MagicMock(return_value=_check_minions_return),
    ):
        # Can we access test.ping?
        await master_acl_clear_funcs.publish(master_acl_valid_load)
        assert (
            master_acl_clear_funcs.event.fire_event.call_args[0][0]["fun"]
            == "test.ping"
        )

        # Are we denied access to sys.doc?
        sys_doc_load = master_acl_valid_load
        sys_doc_load["fun"] = "sys.doc"
        await master_acl_clear_funcs.publish(sys_doc_load)
        assert "error" in master_acl_clear_funcs.event.fire_event.call_args[0][0]


async def test_master_publish_group(master_acl_clear_funcs, master_acl_valid_load):
    """
    Tests to ensure test_group can access test.echo but *not* sys.doc
    """
    _check_minions_return = {"minions": ["some_minions"], "missing": []}
    with patch(
        "salt.utils.minions.CkMinions.check_minions",
        MagicMock(return_value=_check_minions_return),
    ):
        master_acl_valid_load["kwargs"]["user"] = "new_user"
        master_acl_valid_load["fun"] = "test.echo"
        master_acl_valid_load["arg"] = "hello"
        with patch(
            "salt.auth.LoadAuth.get_groups",
            return_value=["test_group", "second_test_group"],
        ):
            await master_acl_clear_funcs.publish(master_acl_valid_load)
        # Did we fire test.echo?
        assert (
            master_acl_clear_funcs.event.fire_event.call_args[0][0]["fun"]
            == "test.echo"
        )

        # Request sys.doc
        master_acl_valid_load["fun"] = "sys.doc"

        # XXX: Of course we won't fire an event if publish isn't called. If
        # sys.dock is there when we publish is that a bug?

        # await master_acl_clear_funcs.publish(master_acl_valid_load)

        # Did we fire it?
        assert (
            master_acl_clear_funcs.event.fire_event.call_args[0][0]["fun"] != "sys.doc"
        )


@pytest.mark.skip_on_windows(reason="PAM eauth not available on Windows")
async def test_master_publish_some_minions(
    master_acl_clear_funcs, master_acl_valid_load
):
    """
    Tests to ensure we can only target minions for which we
    have permission with publisher acl.

    Note that in order for these sorts of tests to run correctly that
    you should NOT patch check_minions!
    """
    master_acl_valid_load["kwargs"]["username"] = "test_user_mminion"
    master_acl_valid_load["user"] = "test_user_mminion"
    await master_acl_clear_funcs.publish(master_acl_valid_load)
    assert master_acl_clear_funcs.event.fire_event.mock_calls == []


async def test_master_not_user_glob_all(master_acl_clear_funcs, master_acl_valid_load):
    """
    Test to ensure that we DO NOT access to a given
    function to all users with publisher acl. ex:

    '*':
        my_minion:
            - my_func

    Yes, this seems like a bit of a no-op test but it's
    here to document that this functionality
    is NOT supported currently.

    WARNING: Do not patch this wit
    """
    master_acl_valid_load["kwargs"]["username"] = "NOT_A_VALID_USERNAME"
    master_acl_valid_load["user"] = "NOT_A_VALID_USERNAME"
    master_acl_valid_load["fun"] = "test.ping"
    await master_acl_clear_funcs.publish(master_acl_valid_load)
    assert "error" in master_acl_clear_funcs.event.fire_event.mock_calls.pop(-1).args[0]
    assert master_acl_clear_funcs.event.fire_event.mock_calls == []


@pytest.mark.skip_on_windows(reason="PAM eauth not available on Windows")
async def test_master_minion_glob(master_acl_clear_funcs, master_acl_valid_load):
    """
    Test to ensure we can allow access to a given
    function for a user to a subset of minions
    selected by a glob. ex:

    test_user:
        'minion_glob*':
          - glob_mod.glob_func

    This test is a bit tricky, because ultimately the real functionality
    lies in what's returned from check_minions, but this checks a limited
    amount of logic on the way there as well. Note the inline patch.
    """
    requested_function = "foo.bar"
    requested_tgt = "minion_glob1"
    master_acl_valid_load["tgt"] = requested_tgt
    master_acl_valid_load["fun"] = requested_function
    _check_minions_return = {"minions": ["minion_glob1"], "missing": []}
    with patch(
        "salt.utils.minions.CkMinions.check_minions",
        MagicMock(return_value=_check_minions_return),
    ):  # Assume that there is a listening minion match
        await master_acl_clear_funcs.publish(master_acl_valid_load)
    assert (
        master_acl_clear_funcs.event.fire_event.called is True
    ), f"Did not fire {requested_function} for minion tgt {requested_tgt}"
    assert (
        master_acl_clear_funcs.event.fire_event.call_args[0][0]["fun"]
        == requested_function
    ), f"Did not fire {requested_function} for minion glob"


@pytest.mark.skip_on_windows(reason="PAM eauth not available on Windows")
async def test_args_empty_spec(master_acl_clear_funcs, master_acl_valid_load):
    """
    Test simple arg restriction allowed.

    'test_user_func':
        minion1:
            - test.empty:
    """
    _check_minions_return = {"minions": ["minion1"], "missing": []}
    with patch(
        "salt.utils.minions.CkMinions.check_minions",
        MagicMock(return_value=_check_minions_return),
    ):
        master_acl_valid_load["kwargs"].update({"username": "test_user_func"})
        master_acl_valid_load.update(
            {
                "user": "test_user_func",
                "tgt": "minion1",
                "fun": "test.empty",
                "arg": ["TEST"],
            }
        )
        await master_acl_clear_funcs.publish(master_acl_valid_load)
        assert (
            master_acl_clear_funcs.event.fire_event.call_args[0][0]["fun"]
            == "test.empty"
        )


@pytest.mark.skip_on_windows(reason="PAM eauth not available on Windows")
async def test_args_simple_match(master_acl_clear_funcs, master_acl_valid_load):
    """
    Test simple arg restriction allowed.

    'test_user_func':
        minion1:
            - test.echo:
                args:
                    - 'TEST'
                    - 'TEST.*'
    """
    _check_minions_return = {"minions": ["minion1"], "missing": []}
    with patch(
        "salt.utils.minions.CkMinions.check_minions",
        MagicMock(return_value=_check_minions_return),
    ):
        master_acl_valid_load["kwargs"].update({"username": "test_user_func"})
        master_acl_valid_load.update(
            {
                "user": "test_user_func",
                "tgt": "minion1",
                "fun": "test.echo",
                "arg": ["TEST", "any", "TEST ABC"],
            }
        )
        await master_acl_clear_funcs.publish(master_acl_valid_load)
        assert (
            master_acl_clear_funcs.event.fire_event.call_args[0][0]["fun"]
            == "test.echo"
        )


@pytest.mark.skip_on_windows(reason="PAM eauth not available on Windows")
async def test_args_more_args(master_acl_clear_funcs, master_acl_valid_load):
    """
    Test simple arg restriction allowed to pass unlisted args.

    'test_user_func':
        minion1:
            - test.echo:
                args:
                    - 'TEST'
                    - 'TEST.*'
    """
    _check_minions_return = {"minions": ["minion1"], "missing": []}
    with patch(
        "salt.utils.minions.CkMinions.check_minions",
        MagicMock(return_value=_check_minions_return),
    ):
        master_acl_valid_load["kwargs"].update({"username": "test_user_func"})
        master_acl_valid_load.update(
            {
                "user": "test_user_func",
                "tgt": "minion1",
                "fun": "test.echo",
                "arg": [
                    "TEST",
                    "any",
                    "TEST ABC",
                    "arg 3",
                    {"kwarg1": "val1", "__kwarg__": True},
                ],
            }
        )
        await master_acl_clear_funcs.publish(master_acl_valid_load)
        assert (
            master_acl_clear_funcs.event.fire_event.call_args[0][0]["fun"]
            == "test.echo"
        )


async def test_args_simple_forbidden(master_acl_clear_funcs, master_acl_valid_load):
    """
    Test simple arg restriction forbidden.

    'test_user_func':
        minion1:
            - test.echo:
                args:
                    - 'TEST'
                    - 'TEST.*'
    """
    _check_minions_return = {"minions": ["minion1"], "missing": []}
    with patch(
        "salt.utils.minions.CkMinions.check_minions",
        MagicMock(return_value=_check_minions_return),
    ):
        master_acl_valid_load["kwargs"].update({"username": "test_user_func"})
        # Wrong last arg
        master_acl_valid_load.update(
            {
                "user": "test_user_func",
                "tgt": "minion1",
                "fun": "test.echo",
                "arg": ["TEST", "any", "TESLA"],
            }
        )
        await master_acl_clear_funcs.publish(master_acl_valid_load)
        assert (
            "error"
            in master_acl_clear_funcs.event.fire_event.mock_calls.pop(-1).args[0]
        )
        assert master_acl_clear_funcs.event.fire_event.mock_calls == []
        # Wrong first arg
        master_acl_valid_load["arg"] = ["TES", "any", "TEST1234"]
        await master_acl_clear_funcs.publish(master_acl_valid_load)
        assert (
            "error"
            in master_acl_clear_funcs.event.fire_event.mock_calls.pop(-1).args[0]
        )
        assert master_acl_clear_funcs.event.fire_event.mock_calls == []
        # Missing the last arg
        master_acl_valid_load["arg"] = ["TEST", "any"]
        await master_acl_clear_funcs.publish(master_acl_valid_load)
        assert (
            "error"
            in master_acl_clear_funcs.event.fire_event.mock_calls.pop(-1).args[0]
        )
        assert master_acl_clear_funcs.event.fire_event.mock_calls == []
        # No args
        master_acl_valid_load["arg"] = []
        await master_acl_clear_funcs.publish(master_acl_valid_load)
        assert (
            "error"
            in master_acl_clear_funcs.event.fire_event.mock_calls.pop(-1).args[0]
        )
        assert master_acl_clear_funcs.event.fire_event.mock_calls == []


@pytest.mark.skip_on_windows(reason="PAM eauth not available on Windows")
async def test_args_kwargs_match(master_acl_clear_funcs, master_acl_valid_load):
    """
    Test simple kwargs restriction allowed.

    'test_user_func':
      '*':
        - test.echo:
            kwargs:
              text: 'KWMSG:.*'
    """
    _check_minions_return = {"minions": ["some_minions"], "missing": []}
    with patch(
        "salt.utils.minions.CkMinions.check_minions",
        MagicMock(return_value=_check_minions_return),
    ):
        master_acl_valid_load["kwargs"].update({"username": "test_user_func"})
        master_acl_valid_load.update(
            {
                "user": "test_user_func",
                "tgt": "*",
                "fun": "test.echo",
                "arg": [
                    {
                        "text": "KWMSG: a message",
                        "anything": "hello all",
                        "none": "hello none",
                        "__kwarg__": True,
                    }
                ],
            }
        )
        await master_acl_clear_funcs.publish(master_acl_valid_load)
        assert (
            master_acl_clear_funcs.event.fire_event.call_args[0][0]["fun"]
            in "test.echo"
        )


async def test_args_kwargs_mismatch(master_acl_clear_funcs, master_acl_valid_load):
    """
    Test simple kwargs restriction allowed.

    'test_user_func':
        '*':
            - test.echo:
                kwargs:
                    text: 'KWMSG:.*'
    """
    _check_minions_return = {"minions": ["some_minions"], "missing": []}
    with patch(
        "salt.utils.minions.CkMinions.check_minions",
        MagicMock(return_value=_check_minions_return),
    ):
        master_acl_valid_load["kwargs"].update({"username": "test_user_func"})
        master_acl_valid_load.update(
            {"user": "test_user_func", "tgt": "*", "fun": "test.echo"}
        )
        # Wrong kwarg value
        master_acl_valid_load["arg"] = [
            {
                "text": "KWMSG a message",
                "anything": "hello all",
                "none": "hello none",
                "__kwarg__": True,
            }
        ]
        await master_acl_clear_funcs.publish(master_acl_valid_load)
        assert (
            "error"
            in master_acl_clear_funcs.event.fire_event.mock_calls.pop(-1).args[0]
        )
        assert master_acl_clear_funcs.event.fire_event.mock_calls == []
        # Missing kwarg value
        master_acl_valid_load["arg"] = [
            {"anything": "hello all", "none": "hello none", "__kwarg__": True}
        ]
        await master_acl_clear_funcs.publish(master_acl_valid_load)
        assert (
            "error"
            in master_acl_clear_funcs.event.fire_event.mock_calls.pop(-1).args[0]
        )
        assert master_acl_clear_funcs.event.fire_event.mock_calls == []
        master_acl_valid_load["arg"] = [{"__kwarg__": True}]
        await master_acl_clear_funcs.publish(master_acl_valid_load)
        assert (
            "error"
            in master_acl_clear_funcs.event.fire_event.mock_calls.pop(-1).args[0]
        )
        assert master_acl_clear_funcs.event.fire_event.mock_calls == []
        master_acl_valid_load["arg"] = [{}]
        await master_acl_clear_funcs.publish(master_acl_valid_load)
        assert (
            "error"
            in master_acl_clear_funcs.event.fire_event.mock_calls.pop(-1).args[0]
        )
        assert master_acl_clear_funcs.event.fire_event.mock_calls == []
        master_acl_valid_load["arg"] = []
        await master_acl_clear_funcs.publish(master_acl_valid_load)
        assert (
            "error"
            in master_acl_clear_funcs.event.fire_event.mock_calls.pop(-1).args[0]
        )
        assert master_acl_clear_funcs.event.fire_event.mock_calls == []
        # Missing kwarg allowing any value
        master_acl_valid_load["arg"] = [
            {"text": "KWMSG: a message", "none": "hello none", "__kwarg__": True}
        ]
        await master_acl_clear_funcs.publish(master_acl_valid_load)
        assert (
            "error"
            in master_acl_clear_funcs.event.fire_event.mock_calls.pop(-1).args[0]
        )
        assert master_acl_clear_funcs.event.fire_event.mock_calls == []
        master_acl_valid_load["arg"] = [
            {"text": "KWMSG: a message", "anything": "hello all", "__kwarg__": True}
        ]
        await master_acl_clear_funcs.publish(master_acl_valid_load)
        assert (
            "error"
            in master_acl_clear_funcs.event.fire_event.mock_calls.pop(-1).args[0]
        )
        assert master_acl_clear_funcs.event.fire_event.mock_calls == []


@pytest.mark.skip_on_windows(reason="PAM eauth not available on Windows")
async def test_args_mixed_match(master_acl_clear_funcs, master_acl_valid_load):
    """
    Test mixed args and kwargs restriction allowed.

    'test_user_func':
        '*':
            - 'my_mod.*':
                args:
                    - 'a.*'
                    - 'b.*'
                kwargs:
                    'kwa': 'kwa.*'
                    'kwb': 'kwb'
    """
    _check_minions_return = {"minions": ["some_minions"], "missing": []}
    with patch(
        "salt.utils.minions.CkMinions.check_minions",
        MagicMock(return_value=_check_minions_return),
    ):
        master_acl_valid_load["kwargs"].update({"username": "test_user_func"})
        master_acl_valid_load.update(
            {
                "user": "test_user_func",
                "tgt": "*",
                "fun": "my_mod.some_func",
                "arg": [
                    "alpha",
                    "beta",
                    "gamma",
                    {
                        "kwa": "kwarg #1",
                        "kwb": "kwb",
                        "one_more": "just one more",
                        "__kwarg__": True,
                    },
                ],
            }
        )
        await master_acl_clear_funcs.publish(master_acl_valid_load)
        assert (
            master_acl_clear_funcs.event.fire_event.call_args[0][0]["fun"]
            == "my_mod.some_func"
        )


async def test_args_mixed_mismatch(master_acl_clear_funcs, master_acl_valid_load):
    """
    Test mixed args and kwargs restriction forbidden.

    'test_user_func':
        '*':
            - 'my_mod.*':
                args:
                    - 'a.*'
                    - 'b.*'
                kwargs:
                    'kwa': 'kwa.*'
                    'kwb': 'kwb'
    """
    _check_minions_return = {"minions": ["some_minions"], "missing": []}
    with patch(
        "salt.utils.minions.CkMinions.check_minions",
        MagicMock(return_value=_check_minions_return),
    ):
        master_acl_valid_load["kwargs"].update({"username": "test_user_func"})
        master_acl_valid_load.update(
            {"user": "test_user_func", "tgt": "*", "fun": "my_mod.some_func"}
        )
        # Wrong arg value
        master_acl_valid_load["arg"] = [
            "alpha",
            "gamma",
            {
                "kwa": "kwarg #1",
                "kwb": "kwb",
                "one_more": "just one more",
                "__kwarg__": True,
            },
        ]
        await master_acl_clear_funcs.publish(master_acl_valid_load)
        assert (
            "error"
            in master_acl_clear_funcs.event.fire_event.mock_calls.pop(-1).args[0]
        )
        assert master_acl_clear_funcs.event.fire_event.mock_calls == []
        # Wrong kwarg value
        master_acl_valid_load["arg"] = [
            "alpha",
            "beta",
            "gamma",
            {
                "kwa": "kkk",
                "kwb": "kwb",
                "one_more": "just one more",
                "__kwarg__": True,
            },
        ]
        await master_acl_clear_funcs.publish(master_acl_valid_load)
        assert (
            "error"
            in master_acl_clear_funcs.event.fire_event.mock_calls.pop(-1).args[0]
        )
        assert master_acl_clear_funcs.event.fire_event.mock_calls == []
        # Missing arg
        master_acl_valid_load["arg"] = [
            "alpha",
            {
                "kwa": "kwarg #1",
                "kwb": "kwb",
                "one_more": "just one more",
                "__kwarg__": True,
            },
        ]
        await master_acl_clear_funcs.publish(master_acl_valid_load)
        assert (
            "error"
            in master_acl_clear_funcs.event.fire_event.mock_calls.pop(-1).args[0]
        )
        assert master_acl_clear_funcs.event.fire_event.mock_calls == []
        # Missing kwarg
        master_acl_valid_load["arg"] = [
            "alpha",
            "beta",
            "gamma",
            {"kwa": "kwarg #1", "one_more": "just one more", "__kwarg__": True},
        ]
        await master_acl_clear_funcs.publish(master_acl_valid_load)
        assert (
            "error"
            in master_acl_clear_funcs.event.fire_event.mock_calls.pop(-1).args[0]
        )
        assert master_acl_clear_funcs.event.fire_event.mock_calls == []


@pytest.mark.skip_on_windows(reason="PAM eauth not available on Windows")
async def test_acl_simple_allow(auth_acl_clear_funcs, auth_acl_valid_load):
    await auth_acl_clear_funcs.publish(auth_acl_valid_load)
    assert auth_acl_clear_funcs.ckminions.auth_check.call_args[0][0] == [
        {"alpha_minion": ["test.ping"]}
    ]


async def test_acl_simple_deny(auth_acl_clear_funcs, auth_acl_valid_load):
    with patch(
        "salt.auth.LoadAuth.get_auth_list",
        MagicMock(return_value=[{"beta_minion": ["test.ping"]}]),
    ):
        await auth_acl_clear_funcs.publish(auth_acl_valid_load)
        assert auth_acl_clear_funcs.ckminions.auth_check.call_args[0][0] == [
            {"beta_minion": ["test.ping"]}
        ]



"""
This is the "master" deltaproxy minion, known better as the `control proxy` because
it controls all the deltaproxies underneath it.
"""

import logging

# This must be present or the Salt loader won"t load this module
__proxyenabled__ = ["deltaproxy"]


# Variables are scoped to this module so we can have persistent data
# across calls to fns in here.
DETAILS = {}

log = logging.getLogger(__file__)


# This does nothing, it"s here just as an example and to provide a log
# entry when the module is loaded.
def __virtual__():
    """
    Only return if all the modules are available
    """
    return True


def init(opts):
    """
    init
    """
    DETAILS["initialized"] = True


def initialized():
    """
    Since grains are loaded in many different places and some of those ws
    places occur before the proxy can be initialized, return whether
    our init() function has been called
    """
    return DETAILS.get("initialized", False)


def grains():
    """
    Make up some grains
    """
    if "grains_cache" not in DETAILS:
        DETAILS["grains_cache"] = {
            "dummy_grain_1": "one",
            "dummy_grain_2": "two",
            "dummy_grain_3": "three",
        }

    return DETAILS["grains_cache"]


def grains_refresh():
    """
    Refresh the grains
    """
    DETAILS["grains_cache"] = None
    return grains()


def ping():
    """
    Degenerate ping
    """
    log.debug("deltaproxy returning ping")
    return True


def shutdown(opts):
    """
    For this proxy shutdown is a no-op
    """
    # TODO call shutdown on all the sub-proxies?
    log.debug("deltaproxy shutdown() called...")



"""
    :codeauthor: Daniel Mizyrycki (mzdaniel@glidelink.net)


    tests.integration.cli.custom_module
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Test salt-ssh sls with a custom module work.

    $ cat srv/custom_module.sls
    custom-module:
      module.run:
        - name: test.recho
        - text: hello


    $ cat srv/_modules/override_test.py
    __virtualname__ = 'test'

    def __virtual__():
        return __virtualname__

    def recho(text):
        return text[::-1]


    $ salt-ssh localhost state.sls custom_module
    localhost:
        olleh
"""

import pytest

from tests.support.case import SSHCase

pytestmark = [
    pytest.mark.skip_on_windows,
    pytest.mark.skipif(
        'grains["osfinger"].startswith(("Fedora Linux-40", "Ubuntu-24.04", "Arch Linux"))',
        reason="System ships with a version of python that is too recent for salt-ssh tests",
        # Actually, the problem is that the tornado we ship is not prepared for Python 3.12,
        # and it imports `ssl` and checks if the `match_hostname` function is defined, which
        # has been deprecated since Python 3.7, so, the logic goes into trying to import
        # backports.ssl-match-hostname which is not installed on the system.
    ),
]


class SSHCustomModuleTest(SSHCase):
    """
    Test sls with custom module functionality using ssh
    """

    @pytest.mark.slow_test
    def test_ssh_regular_module(self):
        """
        Test regular module work using SSHCase environment
        """
        expected = "hello"
        cmd = self.run_function("test.echo", arg=["hello"])
        self.assertEqual(expected, cmd)

    @pytest.mark.slow_test
    @pytest.mark.timeout(120, func_only=True)
    def test_ssh_custom_module(self):
        """
        Test custom module work using SSHCase environment
        """
        expected = "hello"[::-1]
        cmd = self.run_function("test.recho", arg=["hello"])
        self.assertEqual(expected, cmd)

    @pytest.mark.slow_test
    def test_ssh_sls_with_custom_module(self):
        """
        Test sls with custom module work using SSHCase environment
        """
        expected = {
            "module_|-regular-module_|-test.echo_|-run": "hello",
            "module_|-custom-module_|-test.recho_|-run": "olleh",
        }
        cmd = self.run_function("state.sls", arg=["custom_module"])
        for key in cmd:
            if not isinstance(cmd, dict) or not isinstance(cmd[key], dict):
                raise AssertionError(f"{cmd} is not a proper state return")
            elif not cmd[key]["result"]:
                raise AssertionError(cmd[key]["comment"])
            cmd_ret = cmd[key]["changes"].get("ret", None)
            self.assertEqual(cmd_ret, expected[key])



"""
Test the saltcheck module
"""

import pytest

from tests.support.case import ModuleCase


@pytest.mark.timeout_unless_on_windows(120)
class SaltcheckModuleTest(ModuleCase):
    """
    Test the saltcheck module
    """

    @pytest.mark.slow_test
    def test_saltcheck_run(self):
        """
        saltcheck.run_test
        """
        saltcheck_test = {
            "module_and_function": "test.echo",
            "assertion": "assertEqual",
            "expected_return": "This works!",
            "args": ["This works!"],
        }
        ret = self.run_function("saltcheck.run_test", test=saltcheck_test)
        self.assertDictContainsSubset({"status": "Pass"}, ret)

    @pytest.mark.slow_test
    def test_saltcheck_state(self):
        """
        saltcheck.run_state_tests
        """
        saltcheck_test = "validate-saltcheck"
        ret = self.run_function("saltcheck.run_state_tests", [saltcheck_test])
        self.assertDictContainsSubset(
            {"status": "Pass"}, ret[0]["validate-saltcheck"]["echo_test_hello"]
        )
        self.assertDictContainsSubset({"Failed": 0}, ret[1]["TEST RESULTS"])

    @pytest.mark.slow_test
    def test_topfile_validation(self):
        """
        saltcheck.run_highstate_tests
        """
        expected_top_states = self.run_function("state.show_top").get("base", [])
        expected_top_states.append("TEST RESULTS")
        ret = self.run_function("saltcheck.run_highstate_tests")
        for top_state_dict in ret:
            self.assertIn(list(top_state_dict)[0], expected_top_states)

    @pytest.mark.slow_test
    def test_saltcheck_checkall(self):
        """
        Validate saltcheck.run_state_tests check_all for the default saltenv of base.
        validate-saltcheck state hosts a saltcheck-tests directory with 2 .tst files. By running
          check_all=True, both files should be found and show passed results.
        """
        saltcheck_test = "validate-saltcheck"
        ret = self.run_function(
            "saltcheck.run_state_tests", [saltcheck_test], check_all=True
        )
        self.assertDictContainsSubset(
            {"status": "Pass"}, ret[0]["validate-saltcheck"]["echo_test_hello"]
        )
        self.assertDictContainsSubset(
            {"status": "Pass"}, ret[0]["validate-saltcheck"]["check_all_validate"]
        )

    @pytest.mark.slow_test
    def test_saltcheck_checkall_saltenv(self):
        """
        Validate saltcheck.run_state_tests check_all for the prod saltenv
        validate-saltcheck state hosts a saltcheck-tests directory with 2 .tst files. By running
          check_all=True, both files should be found and show passed results.
        """
        saltcheck_test = "validate-saltcheck"
        ret = self.run_function(
            "saltcheck.run_state_tests",
            [saltcheck_test],
            saltenv="prod",
            check_all=True,
        )
        self.assertDictContainsSubset(
            {"status": "Pass"}, ret[0]["validate-saltcheck"]["echo_test_prod_env"]
        )
        self.assertDictContainsSubset(
            {"status": "Pass"}, ret[0]["validate-saltcheck"]["check_all_validate_prod"]
        )

    @pytest.mark.slow_test
    def test_saltcheck_saltenv(self):
        """
        Validate saltcheck.run_state_tests for the prod saltenv
        """
        saltcheck_test = "validate-saltcheck"
        ret = self.run_function(
            "saltcheck.run_state_tests", [saltcheck_test], saltenv="prod"
        )
        self.assertDictContainsSubset(
            {"status": "Pass"}, ret[0]["validate-saltcheck"]["echo_test_prod_env"]
        )



from pathlib import Path

import pytest

from tests.support.runtests import RUNTIME_VARS


@pytest.fixture(scope="module")
def minion_config_overrides():
    return {"master_tops": {"master_tops_test": True}}


@pytest.fixture(scope="module", autouse=True)
def _master_tops_test(state_tree, loaders):
    mod_contents = (
        Path(RUNTIME_VARS.FILES) / "extension_modules" / "tops" / "master_tops_test.py"
    ).read_text()
    try:
        with pytest.helpers.temp_file(
            "master_tops_test.py", mod_contents, state_tree / "_tops"
        ):
            res = loaders.modules.saltutil.sync_tops()
            assert "tops.master_tops_test" in res
            yield
    finally:
        loaders.modules.saltutil.sync_tops()


def test_masterless_master_tops(loaders):
    res = loaders.modules.state.show_top()
    assert res
    assert "base" in res
    assert "master_tops_test" in res["base"]



import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
]


def test_issue_48557(file, tmp_path):
    tempfile = tmp_path / "temp_file_issue_48557"
    contents = "test1\ntest2\ntest3\n"
    expected = "test1\ntest2\ntest4\ntest3\n"
    tempfile.write_text(contents)
    ret = file.line(name=str(tempfile), after="test2", mode="insert", content="test4")
    assert ret.result is True
    assert tempfile.read_text() == expected



"""
Management of MongoDB Databases
===============================

:depends:   - pymongo Python module

Only deletion is supported, creation doesn't make sense and can be done using
:py:func:`mongodb_user.present <salt.states.mongodb_user.present>`.
"""

# Define the module's virtual name
__virtualname__ = "mongodb_database"


def __virtual__():
    if "mongodb.db_exists" in __salt__:
        return __virtualname__
    return (False, "mongodb module could not be loaded")


def absent(name, user=None, password=None, host=None, port=None, authdb=None):
    """
    Ensure that the named database is absent. Note that creation doesn't make
    sense in MongoDB.

    name
        The name of the database to remove

    user
        The user to connect as (must be able to create the user)

    password
        The password of the user

    host
        The host to connect to

    port
        The port to connect to

    authdb
        The database in which to authenticate
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    if __salt__["mongodb.db_exists"](name, user, password, host, port, authdb=authdb):
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "Database {} is present and needs to be removed".format(
                name
            )
            return ret
        if __salt__["mongodb.db_remove"](
            name, user, password, host, port, authdb=authdb
        ):
            ret["comment"] = f"Database {name} has been removed"
            ret["changes"][name] = "Absent"
            return ret

    ret["comment"] = f"Database {name} is not present"
    return ret



import asyncio
import logging
import multiprocessing
import os
import socket
import time
import warnings

import aiohttp
import aiohttp.web
import tornado.ioloop

import salt.payload
import salt.transport.base
import salt.transport.frame
from salt.transport.tcp import (
    USE_LOAD_BALANCER,
    LoadBalancerServer,
    _get_bind_addr,
    _get_socket,
    _set_tcp_keepalive,
)

log = logging.getLogger(__name__)


class PublishClient(salt.transport.base.PublishClient):
    """
    Tornado based TCP Pub Client
    """

    ttype = "ws"

    async_methods = [
        "connect",
        "connect_uri",
        "recv",
        # "close",
    ]
    close_methods = [
        "close",
    ]

    def __init__(self, opts, io_loop, **kwargs):  # pylint: disable=W0231
        self.opts = opts
        self.io_loop = io_loop

        self.connected = False
        self._closing = False
        self._closing = False
        self._closed = False

        self.backoff = opts.get("tcp_reconnect_backoff", 1)
        self.poller = None

        self.host = kwargs.get("host", None)
        self.port = kwargs.get("port", None)
        self.path = kwargs.get("path", None)
        self.ssl = kwargs.get("ssl", None)
        self.source_ip = self.opts.get("source_ip")
        self.source_port = self.opts.get("source_publish_port")
        self.connect_callback = None
        self.disconnect_callback = None
        if self.host is None and self.port is None:
            if self.path is None:
                raise Exception("A host and port or a path must be provided")
        elif self.host and self.port:
            if self.path:
                raise Exception("A host and port or a path must be provided, not both")
        self._ws = None
        self._session = None
        self._closing = False
        self.on_recv_task = None

    async def _close(self):
        if self._session is not None:
            await self._session.close()
            self._session = None
        if self.on_recv_task:
            self.on_recv_task.cancel()
            await self.on_recv_task
            self.on_recv_task = None
        if self._ws is not None:
            await self._ws.close()
            self._ws = None
        self._closed = True

    def close(self):
        if self._closing:
            return
        self._closing = True
        self.io_loop.spawn_callback(self._close)

    # pylint: disable=W1701
    def __del__(self):
        if not self._closing:
            warnings.warn(
                "unclosed publish client {self!r}", ResourceWarning, source=self
            )

    # pylint: enable=W1701
    def _decode_messages(self, messages):
        if not isinstance(messages, dict):
            body = salt.payload.loads(messages)
        else:
            body = messages
        return body

    async def getstream(self, **kwargs):
        if self.source_ip or self.source_port:
            kwargs.update(source_ip=self.source_ip, source_port=self.source_port)
        ws = None
        start = time.monotonic()
        timeout = kwargs.get("timeout", None)
        while ws is None and (not self._closed and not self._closing):
            session = None
            try:
                ctx = None
                if self.ssl is not None:
                    ctx = salt.transport.base.ssl_context(self.ssl, server_side=False)
                if self.host and self.port:
                    conn = aiohttp.TCPConnector()
                    session = aiohttp.ClientSession(connector=conn)
                    if self.ssl:
                        url = f"https://{self.host}:{self.port}/ws"
                    else:
                        url = f"http://{self.host}:{self.port}/ws"
                else:
                    conn = aiohttp.UnixConnector(path=self.path)
                    session = aiohttp.ClientSession(connector=conn)
                    if self.ssl:
                        url = "https://ipc.saltproject.io/ws"
                    else:
                        url = "http://ipc.saltproject.io/ws"
                log.debug("pub client connect %r %r", url, ctx)
                ws = await asyncio.wait_for(session.ws_connect(url, ssl=ctx), 3)
            except Exception as exc:  # pylint: disable=broad-except
                log.warning(
                    "WS Message Client encountered an exception while connecting to"
                    " %s:%s %s: %r, will reconnect in %d seconds",
                    self.host,
                    self.port,
                    self.path,
                    exc,
                    self.backoff,
                )
                if session:
                    await session.close()
                if timeout and time.monotonic() - start > timeout:
                    break
                await asyncio.sleep(self.backoff)
        return ws, session

    async def _connect(self, timeout=None):
        if self._ws is None:
            self._ws, self._session = await self.getstream(timeout=timeout)
            if self.connect_callback:
                self.connect_callback(True)  # pylint: disable=not-callable
            self.connected = True

    async def connect(
        self,
        port=None,
        connect_callback=None,
        disconnect_callback=None,
        timeout=None,
    ):
        if port is not None:
            self.port = port
        if connect_callback:
            self.connect_callback = None
        if disconnect_callback:
            self.disconnect_callback = None
        await self._connect(timeout=timeout)

    async def send(self, msg):
        await self.message_client.send(msg, reply=False)

    async def recv(self, timeout=None):
        while self._ws is None:
            await self.connect()
            await asyncio.sleep(0.001)
        if timeout == 0:
            try:
                raw_msg = await asyncio.wait_for(self._ws.receive(), 0.0001)
            except TimeoutError:
                return
            if raw_msg.type == aiohttp.WSMsgType.TEXT:
                if raw_msg.data == "close":
                    await self._ws.close()
            if raw_msg.type == aiohttp.WSMsgType.BINARY:
                return salt.payload.loads(raw_msg.data, raw=True)
            elif raw_msg.type == aiohttp.WSMsgType.ERROR:
                log.error(
                    "ws connection closed with exception %s", self._ws.exception()
                )
        elif timeout:
            return await asyncio.wait_for(self.recv(), timeout=timeout)
        else:
            while True:
                raw_msg = await self._ws.receive()
                if raw_msg.type == aiohttp.WSMsgType.TEXT:
                    if raw_msg.data == "close":
                        await self._ws.close()
                if raw_msg.type == aiohttp.WSMsgType.BINARY:
                    return salt.payload.loads(raw_msg.data, raw=True)
                elif raw_msg.type == aiohttp.WSMsgType.ERROR:
                    log.error(
                        "ws connection closed with exception %s",
                        self._ws.exception(),
                    )

    async def on_recv_handler(self, callback):
        while not self._ws:
            await asyncio.sleep(0.003)
        while True:
            msg = await self.recv()
            await callback(msg)

    def on_recv(self, callback):
        """
        Register a callback for received messages (that we didn't initiate)
        """
        if self.on_recv_task:
            # XXX: We are not awaiting this canceled task. This still needs to
            # be addressed.
            self.on_recv_task.cancel()
        if callback is None:
            self.on_recv_task = None
        else:
            self.on_recv_task = asyncio.create_task(self.on_recv_handler(callback))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class PublishServer(salt.transport.base.DaemonizedPublishServer):
    """ """

    # TODO: opts!
    # Based on default used in tornado.netutil.bind_sockets()
    backlog = 128
    async_methods = [
        "publish",
        # "close",
    ]
    close_methods = [
        "close",
    ]

    def __init__(
        self,
        opts,
        pub_host=None,
        pub_port=None,
        pub_path=None,
        pull_host=None,
        pull_port=None,
        pull_path=None,
        pull_path_perms=0o600,
        pub_path_perms=0o600,
        ssl=None,
        started=None,
    ):
        self.opts = opts
        self.pub_host = pub_host
        self.pub_port = pub_port
        self.pub_path = pub_path
        self.pull_host = pull_host
        self.pull_port = pull_port
        self.pull_path = pull_path
        self.pull_path_perms = pull_path_perms
        self.pub_path_perms = pub_path_perms
        self.ssl = ssl
        self.clients = set()
        self._run = None
        self.pub_writer = None
        self.pub_reader = None
        self._connecting = None
        if started is None:
            self.started = multiprocessing.Event()
        else:
            self.started = started

    @property
    def topic_support(self):
        return not self.opts.get("order_masters", False)

    def __setstate__(self, state):
        self.__init__(**state)

    def __getstate__(self):
        return {
            "opts": self.opts,
            "pub_host": self.pub_host,
            "pub_port": self.pub_port,
            "pub_path": self.pub_path,
            "pull_host": self.pull_host,
            "pull_port": self.pull_port,
            "pull_path": self.pull_path,
            "pull_path_perms": self.pull_path_perms,
            "pub_path_perms": self.pub_path_perms,
            "ssl": self.ssl,
            "started": self.started,
        }

    def publish_daemon(
        self,
        publish_payload,
        presence_callback=None,
        remove_presence_callback=None,
    ):
        """
        Bind to the interface specified in the configuration file
        """
        io_loop = tornado.ioloop.IOLoop()
        io_loop.add_callback(
            self.publisher,
            publish_payload,
            presence_callback,
            remove_presence_callback,
            io_loop,
        )
        # run forever
        try:
            io_loop.start()
        except (KeyboardInterrupt, SystemExit):
            pass
        finally:
            self.close()

    async def publisher(
        self,
        publish_payload,
        presence_callback=None,
        remove_presence_callback=None,
        io_loop=None,
    ):
        if io_loop is None:
            io_loop = tornado.ioloop.IOLoop.current()
        if self._run is None:
            self._run = asyncio.Event()
        self._run.set()

        ctx = None
        if self.ssl is not None:
            ctx = salt.transport.base.ssl_context(self.ssl, server_side=True)
        if self.pub_path:
            server = aiohttp.web.Server(self.handle_request)
            runner = aiohttp.web.ServerRunner(server)
            await runner.setup()
            with salt.utils.files.set_umask(0o177):
                log.info("Publisher binding to socket %s", self.pub_path)
                site = aiohttp.web.UnixSite(runner, self.pub_path, ssl_context=ctx)
                await site.start()
                os.chmod(self.pub_path, self.pub_path_perms)
        else:
            sock = _get_socket(self.opts)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            # _set_tcp_keepalive(sock, self.opts)
            sock.setblocking(0)
            sock.bind((self.pub_host, self.pub_port))
            sock.listen(self.backlog)
            server = aiohttp.web.Server(self.handle_request)
            runner = aiohttp.web.ServerRunner(server)
            await runner.setup()
            site = aiohttp.web.SockSite(runner, sock, ssl_context=ctx)
            log.info("Publisher binding to socket %s:%s", self.pub_host, self.pub_port)
            await site.start()

        self._pub_payload = publish_payload
        if self.pull_path:
            with salt.utils.files.set_umask(0o177):
                self.puller = await asyncio.start_unix_server(
                    self.pull_handler, self.pull_path
                )
                os.chmod(self.pull_path, self.pull_path_perms)
        else:
            self.puller = await asyncio.start_server(
                self.pull_handler, self.pull_host, self.pull_port
            )
        self.started.set()
        while self._run.is_set():
            await asyncio.sleep(0.3)
        await self.server.stop()
        await self.puller.wait_closed()

    async def pull_handler(self, reader, writer):
        unpacker = salt.utils.msgpack.Unpacker(raw=True)
        while True:
            data = await reader.read(1024)
            unpacker.feed(data)
            for msg in unpacker:
                await self._pub_payload(msg)

    def pre_fork(self, process_manager):
        """
        Do anything necessary pre-fork. Since this is on the master side this will
        primarily be used to create IPC channels and create our daemon process to
        do the actual publishing
        """
        process_manager.add_process(
            self.publish_daemon,
            args=[self.publish_payload],
            name=self.__class__.__name__,
        )

    async def handle_request(self, request):
        try:
            cert = request.get_extra_info("peercert")
        except AttributeError:
            pass
        else:
            if cert:
                name = salt.transport.base.common_name(cert)
                log.debug("Request client cert %r", name)
        ws = aiohttp.web.WebSocketResponse()
        await ws.prepare(request)
        self.clients.add(ws)
        while True:
            await asyncio.sleep(1)

    async def _connect(self):
        if self.pull_path:
            self.pub_reader, self.pub_writer = await asyncio.open_unix_connection(
                self.pull_path
            )
        else:
            self.pub_reader, self.pub_writer = await asyncio.open_connection(
                self.pull_host, self.pull_port
            )
        self._connecting = None

    def connect(self):
        log.debug("Connect pusher %s", self.pull_path)
        if self._connecting is None:
            self._connecting = asyncio.create_task(self._connect())
        return self._connecting

    async def publish(
        self, payload, **kwargs
    ):  # pylint: disable=invalid-overridden-method
        """
        Publish "load" to minions
        """
        if not self.pub_writer:
            await self.connect()
        self.pub_writer.write(salt.payload.dumps(payload, use_bin_type=True))
        await self.pub_writer.drain()

    async def publish_payload(self, package, *args):
        payload = salt.payload.dumps(package, use_bin_type=True)
        for ws in list(self.clients):
            try:
                await ws.send_bytes(payload)
            except ConnectionResetError:
                self.clients.discard(ws)

    def close(self):
        if self.pub_writer:
            self.pub_writer.close()
            self.pub_writer = None
            self.pub_reader = None
        if self._run is not None:
            self._run.clear()
        if self._connecting:
            self._connecting.cancel()


class RequestServer(salt.transport.base.DaemonizedRequestServer):
    def __init__(self, opts):  # pylint: disable=W0231
        self.opts = opts
        self.site = None
        self.ssl = self.opts.get("ssl", None)

    def pre_fork(self, process_manager):
        """
        Pre-fork we need to create the zmq router device
        """
        if USE_LOAD_BALANCER:
            self.socket_queue = multiprocessing.Queue()
            process_manager.add_process(
                LoadBalancerServer,
                args=(self.opts, self.socket_queue),
                name="LoadBalancerServer",
            )
        elif not salt.utils.platform.is_windows():
            self._socket = _get_socket(self.opts)
            self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            _set_tcp_keepalive(self._socket, self.opts)
            self._socket.setblocking(0)
            self._socket.bind(_get_bind_addr(self.opts, "ret_port"))

    def post_fork(self, message_handler, io_loop):
        """
        After forking we need to create all of the local sockets to listen to the
        router

        message_handler: function to call with your payloads
        """
        self.message_handler = message_handler
        self._run = asyncio.Event()
        self._started = asyncio.Event()
        self._run.set()

        async def server():
            server = aiohttp.web.Server(self.handle_message)
            runner = aiohttp.web.ServerRunner(server)
            await runner.setup()
            ctx = None
            if self.ssl is not None:
                ctx = tornado.netutil.ssl_options_to_context(self.ssl, server_side=True)
            self.site = aiohttp.web.SockSite(runner, self._socket, ssl_context=ctx)
            log.info("Worker binding to socket %s", self._socket)
            await self.site.start()
            self._started.set()
            # pause here for very long time by serving HTTP requests and
            # waiting for keyboard interruption
            while self._run.is_set():
                await asyncio.sleep(0.3)
            await self.site.stop()

        io_loop.spawn_callback(server)

    async def handle_message(self, request):
        try:
            cert = request.get_extra_info("peercert")
        except AttributeError:
            pass
        else:
            if cert:
                name = salt.transport.base.common_name(cert)
                log.debug("Request client cert %r", name)
        ws = aiohttp.web.WebSocketResponse()
        await ws.prepare(request)
        async for msg in ws:
            if msg.type == aiohttp.WSMsgType.TEXT:
                if msg.data == "close":
                    await ws.close()
            if msg.type == aiohttp.WSMsgType.BINARY:
                payload = salt.payload.loads(msg.data)
                reply = await self.message_handler(payload)
                await ws.send_bytes(salt.payload.dumps(reply))
            elif msg.type == aiohttp.WSMsgType.ERROR:
                log.error("ws connection closed with exception %s", ws.exception())

    def close(self):
        self._run.clear()


class RequestClient(salt.transport.base.RequestClient):

    ttype = "ws"

    def __init__(self, opts, io_loop):  # pylint: disable=W0231
        self.opts = opts
        self.sending = False
        self.ws = None
        self.session = None
        self.io_loop = io_loop
        self._closing = False
        self._closed = False
        self.ssl = self.opts.get("ssl", None)

    async def connect(self):  # pylint: disable=invalid-overridden-method
        ctx = None
        if self.ssl is not None:
            ctx = tornado.netutil.ssl_options_to_context(self.ssl, server_side=False)
        self.session = aiohttp.ClientSession()
        URL = self.get_master_uri(self.opts)
        log.debug("Connect to %s %s", URL, ctx)
        self.ws = await self.session.ws_connect(URL, ssl=ctx)

    async def send(self, load, timeout=60):
        if self.sending or self._closing:
            await asyncio.sleep(0.03)
        self.sending = True
        try:
            await self.connect()
            message = salt.payload.dumps(load)
            await self.ws.send_bytes(message)
            async for msg in self.ws:
                if msg.type in (aiohttp.WSMsgType.CLOSED, aiohttp.WSMsgType.ERROR):
                    break
                data = salt.payload.loads(msg.data)
                break
            return data
        finally:
            self.sending = False

    async def _close(self):
        if self.ws is not None:
            await self.ws.close()
            self.ws = None
        if self.session is not None:
            await self.session.close()
            self.session = None
        self._closed = True

    def close(self):
        if self._closing:
            return
        self._closing = True
        self.close_task = asyncio.create_task(self._close())

    def get_master_uri(self, opts):
        if "master_uri" in opts:
            if self.opts.get("ssl", None):
                return opts["master_uri"].replace("tcp:", "https:", 1)
            return opts["master_uri"].replace("tcp:", "http:", 1)
        if self.opts.get("ssl", None):
            return f"https://{opts['master_ip']}:{opts['master_port']}/ws"
        return f"http://{opts['master_ip']}:{opts['master_port']}/ws"

    # pylint: disable=W1701
    def __del__(self):
        if not self._closing:
            warnings.warn(
                "Unclosed publish client {self!r}", ResourceWarning, source=self
            )

    # pylint: enable=W1701



"""
Test cases for salt.states.openvswitch_bridge.
"""

import pytest

import salt.states.openvswitch_bridge as openvswitch_bridge
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {openvswitch_bridge: {"__opts__": {"test": False}}}


def test_present_no_parent_existing_no_parent():
    """
    Test present function, not specifying a parent.

    This tests the case where the bridge already exists and has no parent.
    """
    create_mock = MagicMock()
    exists_mock = MagicMock(return_value=True)
    to_parent_mock = MagicMock(return_value="br0")
    to_vlan_mock = MagicMock(return_value=0)
    with patch.dict(
        openvswitch_bridge.__salt__,
        {
            "openvswitch.bridge_create": create_mock,
            "openvswitch.bridge_exists": exists_mock,
            "openvswitch.bridge_to_parent": to_parent_mock,
            "openvswitch.bridge_to_vlan": to_vlan_mock,
        },
    ):
        ret = openvswitch_bridge.present(name="br0")
        create_mock.assert_not_called()
        assert ret["result"] is True


def test_present_no_parent_existing_with_parent():
    """
    Test present function, not specifying a parent.

    This tests the case where the bridge already exists and has a parent.
    """
    create_mock = MagicMock()
    exists_mock = MagicMock(return_value=True)
    to_parent_mock = MagicMock(return_value="br0")
    to_vlan_mock = MagicMock(return_value=42)
    with patch.dict(
        openvswitch_bridge.__salt__,
        {
            "openvswitch.bridge_create": create_mock,
            "openvswitch.bridge_exists": exists_mock,
            "openvswitch.bridge_to_parent": to_parent_mock,
            "openvswitch.bridge_to_vlan": to_vlan_mock,
        },
    ):
        # Bridge exists, but parent and VLAN do not match, so we expect a
        # result of False.
        ret = openvswitch_bridge.present(name="br1")
        create_mock.assert_not_called()
        assert ret["result"] is False


def test_present_no_parent_not_existing():
    """
    Test present function, not specifying a parent.

    This tests the case where the bridge does not exist yet.
    """
    create_mock = MagicMock(return_value=True)
    exists_mock = MagicMock(return_value=False)
    with patch.dict(
        openvswitch_bridge.__salt__,
        {
            "openvswitch.bridge_create": create_mock,
            "openvswitch.bridge_exists": exists_mock,
        },
    ):
        ret = openvswitch_bridge.present(name="br0")
        create_mock.assert_called_with("br0", parent=None, vlan=None)
        assert ret["result"] is True
        assert ret["changes"] == {
            "br0": {"new": "Bridge br0 created", "old": "Bridge br0 does not exist."}
        }


def test_present_with_parent_existing_with_parent():
    """
    Test present function, specifying a parent.

    This tests the case where the bridge already exists and has a parent that
    matches the specified one.
    """
    create_mock = MagicMock()
    exists_mock = MagicMock(return_value=True)
    to_parent_mock = MagicMock(return_value="br0")
    to_vlan_mock = MagicMock(return_value=42)
    with patch.dict(
        openvswitch_bridge.__salt__,
        {
            "openvswitch.bridge_create": create_mock,
            "openvswitch.bridge_exists": exists_mock,
            "openvswitch.bridge_to_parent": to_parent_mock,
            "openvswitch.bridge_to_vlan": to_vlan_mock,
        },
    ):
        # Bridge exists and parent VLAN matches
        ret = openvswitch_bridge.present(name="br1", parent="br0", vlan=42)
        create_mock.assert_not_called()
        assert ret["result"] is True


def test_present_with_parent_not_existing():
    """
    Test present function, specifying a parent.

    This tests the case where the bridge does not exist yet.
    """
    create_mock = MagicMock(return_value=True)
    exists_mock = MagicMock(return_value=False)
    with patch.dict(
        openvswitch_bridge.__salt__,
        {
            "openvswitch.bridge_create": create_mock,
            "openvswitch.bridge_exists": exists_mock,
        },
    ):
        ret = openvswitch_bridge.present(name="br1", parent="br0", vlan=42)
        create_mock.assert_called_with("br1", parent="br0", vlan=42)
        assert ret["result"] is True
        assert ret["changes"] == {
            "br1": {"new": "Bridge br1 created", "old": "Bridge br1 does not exist."}
        }
